# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


__name__      = 'STochastic Engine for Pathway Simulation'
__version__   = '1.1.4'
__author__    = 'STEPS Development Team'
__url__       = 'steps.sourceforge.net'
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# END



# -*- coding: utf-8 -*-

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#  Last Changed Rev:  $Rev: 410 $
#  Last Changed Date: $Date: 2011-04-07 16:11:28 +0900 (Thu, 07 Apr 2011) $
#  Last Changed By:   $Author: iain $

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# This file is the user-interface file for all solver objects.
# All objects are directly derived from the corresponding swig objects.
# All objects are owned by Python.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
"""
Implementation of simulation solvers. 

Each solver is a partial or full implementation of the STEPS solver API.
At the moment STEPS implements three different solvers. 

The steps.solver.Wmrk4 class implements a well-mixed, deterministic solver 
based on the Runge–Kutta method. 

The steps.solver.Wmdirect class implements a stochastic, well-mixed solver 
based on Gillespie's Direct SSA Method. 

The steps.solver.Tetexact class implements a stochastic reaction-diffusion 
solver, based on Gillespie's Direct SSA Method extended for diffusive fluxes 
between tetrahedral elements in complex geometries.

"""
from . import steps_swig
import _steps_swig
import cPickle

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Well-mixed RK4
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
class Wmrk4(steps_swig.Wmrk4) :
    def __init__(self, model, geom, rng): 
        """
        Construction::
        
            sim = steps.solver.Wmrk4(model, geom, rng)
            
        Create a well-mixed RK4 simulation solver.
            
        Arguments: 
            * steps.model.Model model
            * steps.geom.Geom geom
            * steps.rng.RNG rng
        """
        this = _steps_swig.new_Wmrk4(model, geom, rng)
        try: self.this.append(this)
        except: self.this = this
        self.thisown = True
        self.model = model
        self.geom = geom
        
    def run(self, end_time, cp_interval = 0.0, prefix = ""):
        """
        Run the simulation until end_time, 
        automatically checkpoint at each cp_interval.
        Prefix can be added using prefix=<prefix_string>.
        """
        
        if cp_interval > 0:
            while _steps_swig.API_getTime(self) + cp_interval < end_time:
                _steps_swig.API_advance(self, cp_interval)
                filename = "%s%e.wmrk4_cp" % (prefix, _steps_swig.API_getTime(self))
                print "Checkpointing -> ", filename
                _steps_swig.API_checkpoint(self, filename)
            _steps_swig.API_run(self, end_time)
            filename = "%s%e.wmrk4_cp" % (prefix, _steps_swig.API_getTime(self))
            print "Checkpointing -> ", filename
            _steps_swig.API_checkpoint(self, filename)
        else:
            _steps_swig.API_run(self, end_time)
        
    def advance(self, advance_time, cp_interval = 0.0, prefix = ""):
        """
        Advance the simulation for advance_time, 
        automatically checkpoint at each cp_interval.
        Prefix can be added using prefix=<prefix_string>.
        """
        
        end_time = _steps_swig.API_getTime(self) + advance_time
        if cp_interval > 0:
            while _steps_swig.API_getTime(self) + cp_interval < end_time:
                _steps_swig.API_advance(self, cp_interval)
                filename = "%s%e.wmrk4_cp" % (prefix, _steps_swig.API_getTime(self))
                print "Checkpointing -> ", filename
                _steps_swig.API_checkpoint(self, filename)
            _steps_swig.API_run(self, end_time)
            filename = "%s%e.wmrk4_cp" % (prefix, _steps_swig.API_getTime(self))
            print "Checkpointing -> ", filename
            _steps_swig.API_checkpoint(self, filename)
        else:
            _steps_swig.API_run(self, end_time)
            

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Well-mixed Direct SSA
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
class Wmdirect(steps_swig.Wmdirect) :
    def __init__(self, model, geom, rng): 
        """
        Construction::
        
            sim = steps.solver.Wmdirect(model, geom, rng)
            
        Create a well-mixed Direct SSA simulation solver.
            
        Arguments: 
            * steps.model.Model model
            * steps.geom.Geom geom
            * steps.rng.RNG rng
        """
        this = _steps_swig.new_Wmdirect(model, geom, rng)
        try: self.this.append(this)
        except: self.this = this
        self.thisown = True
        self.model = model
        self.geom = geom
        
    def run(self, end_time, cp_interval = 0.0, prefix = ""):
        """
        Run the simulation until <end_time>, 
        automatically checkpoint at each <cp_interval>.
        Prefix can be added using prefix=<prefix_string>.
        """
        
        if cp_interval > 0:
            while _steps_swig.API_getTime(self) + cp_interval < end_time:
                _steps_swig.API_advance(self, cp_interval)
                filename = "%s%e.wmdirect_cp" % (prefix, _steps_swig.API_getTime(self))
                print "Checkpointing -> ", filename
                _steps_swig.API_checkpoint(self, filename)
            _steps_swig.API_run(self, end_time)
            filename = "%s%e.wmdirect_cp" % (prefix, _steps_swig.API_getTime(self))
            print "Checkpointing -> ", filename
            _steps_swig.API_checkpoint(self, filename)
        else:
            _steps_swig.API_run(self, end_time)
        
    def advance(self, advance_time, cp_interval = 0.0):
        """
        Advance the simulation for advance_time, 
        automatically checkpoint at each cp_interval.
        Prefix can be added using prefix=<prefix_string>.
        """
        
        end_time = _steps_swig.API_getTime(self) + advance_time
        if cp_interval > 0:
            while _steps_swig.API_getTime(self) + cp_interval < end_time:
                _steps_swig.API_advance(self, cp_interval)
                filename = "%s%e.wmdirect_cp" % (prefix, _steps_swig.API_getTime(self))
                print "Checkpointing -> ", filename
                _steps_swig.API_checkpoint(self, filename)
            _steps_swig.API_run(self, end_time)
            filename = "%s%e.wmdirect_cp" % (prefix, _steps_swig.API_getTime(self))
            print "Checkpointing -> ", filename
            _steps_swig.API_checkpoint(self, filename)
        else:
            _steps_swig.API_run(self, end_time)
            
        
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Tetrahedral Direct SSA
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #        
class Tetexact(steps_swig.Tetexact) :  
    def __init__(self, model, geom, rng): 
        """
        Construction::
        
            sim = steps.solver.Tetexact(model, geom, rng)
            
        Create a Tetexact SSA simulation solver.
            
        Arguments: 
            * steps.model.Model model
            * steps.geom.Geom geom
            * steps.rng.RNG rng
        """
        this = _steps_swig.new_Tetexact(model, geom, rng)
        try: self.this.append(this)
        except: self.this = this
        self.thisown = True
        self.model = model
        self.geom = geom

        
    def run(self, end_time, cp_interval = 0.0, prefix = ""):
        """
        Run the simulation until <end_time>, 
        automatically checkpoint at each <cp_interval>.
        Prefix can be added using prefix=<prefix_string>.
        """
        
        if cp_interval > 0:
            while _steps_swig.API_getTime(self) + cp_interval < end_time:
                _steps_swig.API_advance(self, cp_interval)
                filename = "%s%e.tetexact_cp" % (prefix, _steps_swig.API_getTime(self))
                print "Checkpointing -> ", filename
                _steps_swig.API_checkpoint(self, filename)
            _steps_swig.API_run(self, end_time)
            filename = "%s%e.tetexact_cp" % (prefix, _steps_swig.API_getTime(self))
            print "Checkpointing -> ", filename
            _steps_swig.API_checkpoint(self, filename)
        else:
            _steps_swig.API_run(self, end_time)
        
    def advance(self, advance_time, cp_interval = 0.0, prefix = ""):
        """
        Advance the simulation for <advance_time>, 
        automatically checkpoint at each <cp_interval>.
        Prefix can be added using prefix=<prefix_string>.
        """
        
        end_time = _steps_swig.API_getTime(self) + advance_time
        if cp_interval > 0:
            while _steps_swig.API_getTime(self) + cp_interval < end_time:
                _steps_swig.API_advance(self, cp_interval)
                filename = "%s%e.tetexact_cp" % (prefix, _steps_swig.API_getTime(self))
                print "Checkpointing -> ", filename
                _steps_swig.API_checkpoint(self, filename)
            _steps_swig.API_run(self, end_time)
            filename = "%s%e.tetexact_cp" % (prefix, _steps_swig.API_getTime(self))
            print "Checkpointing -> ", filename
            _steps_swig.API_checkpoint(self, filename)
        else:
            _steps_swig.API_run(self, end_time)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Example: Unbounded diffusion
# http://steps.sourceforge.net/manual/diffusion.html

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import math
import numpy
import pylab
import random
import time

import steps.model as smodel
import steps.solver as solvmod
import steps.geom as stetmesh
import steps.rng as srng

import steps.utilities.meshio as smeshio

########################################################################

# The number of iterations to run 
NITER = 10
# The data collection time increment (s)
DT = 0.001
# The simulation endtime (s)
INT = 0.101

# The number of molecules to be injected into the centre
NINJECT = 10000	

# The number of tetrahedral elements to sample data from. 
SAMPLE = 2000	

# The diffusion constant for our diffusing species (m^2/s)
DCST= 20.0e-12

# Array to hold tetrahedron indices (integers)
tetidxs = numpy.zeros(SAMPLE, dtype = 'int')

# Array to hold tetrahedron radial distances from mesh centre (floats) 
tetrads = numpy.zeros(SAMPLE)

beg_time = time.time()

########################################################################

def printtime(end_time):

	totsecs = int(end_time-beg_time)
	sec = totsecs%60
	totmin = totsecs/60
	min = totmin%60
	hours = totmin/60
	
	print 'Simulation time: %d h, %d min, %d sec' %(hours, min, sec)
	
########################################################################

def gen_model():         
	mdl = smodel.Model()        
	A = smodel.Spec('A', mdl)                        
	vsys = smodel.Volsys('cytosolv', mdl)
	diff_A = smodel.Diff('diff_A', vsys, A, dcst = DCST)
	return mdl
	
########################################################################

def gen_geom():
    
    print "Loading mesh..."
    mesh = smeshio.loadMesh('meshes/sphere_rad10_11Ktets')[0]
    print "Mesh Loaded"
    
    # Find the total number of tetrahedrons in the mesh	
    ntets = mesh.countTets()
    # Create a compartment containing all tetrahedron
    comp = stetmesh.TmComp('cyto', mesh, range(ntets))
    comp.addVolsys('cytosolv')
    
    print "Finding tetrahedron samples..."
    # Fetch the central tetrahedron index and store:
    ctetidx = mesh.findTetByPoint([0.0, 0.0, 0.0])
    tetidxs[0] = ctetidx

    # Find the central tetrahedron's four neighbours:
    neighbidcs = mesh.getTetTetNeighb(ctetidx)
    tetidxs[1], tetidxs[2], tetidxs[3], tetidxs[4] = neighbidcs
    
    # Keep track how many tet indices we have stored so far
    stored = 5
    
    # Find the maximum and minimum coordinates of the mesh
    max = mesh.getBoundMax()
    min = mesh.getBoundMin()
    
    # Run a loop until we have stored all tet indices we require
    while (stored < SAMPLE):

        # Fetch 3 random numbers between 0 and 1
        rnx = random.random()
        rny = random.random()
        rnz = random.random()  
              
        # Find the coordinates in the mesh that these numbers relate to
        xcrd = min[0] + (max[0]-min[0])*rnx
        ycrd = min[1] + (max[1]-min[1])*rny
        zcrd = min[2] + (max[2]-min[2])*rnz
        
        # Find the tetrahedron that encompasses this point.
        tidx = mesh.findTetByPoint([xcrd, ycrd, zcrd])
        
        # -1 was returned if point is outside the mesh:
        if (tidx == -1): continue
        if (tidx not in tetidxs):
            tetidxs[stored] = tidx
            stored += 1
    
    # Find the barycenter of the central tetrahedron
    cbaryc = mesh.getTetBarycenter(ctetidx)
    
    for i in range(SAMPLE):
        # Fetch the barycenter of the tetrahedron:
        baryc = mesh.getTetBarycenter(tetidxs[i])

        # Find the radial distance of this tetrahedron to mesh center:
        r = math.sqrt(math.pow((baryc[0]-cbaryc[0]),2) + \
                math.pow((baryc[1]-cbaryc[1]),2) + \
                    math.pow((baryc[2]-cbaryc[2]),2))

        # Store the radial distance (in microns):
        tetrads[i] = r*1.0e6
        
    print "Tetrahedron samples found"
    
    return mesh
    
########################################################################

model = gen_model()
tmgeom = gen_geom()

rng = srng.create('mt19937', 512) 
rng.initialize(2903) # The max unsigned long

sim = solvmod.Tetexact(model, tmgeom, rng)

tpnts = numpy.arange(0.0, INT, DT)
# Find how many "time points" we have
ntpnts = tpnts.shape[0]

# Create the data structure: iterations x time points x tet samples
res = numpy.zeros((NITER, ntpnts, SAMPLE))

# Fetch the index of the tetrahedron at the centre of the mesh
ctetidx = tmgeom.findTetByPoint([0.0, 0.0, 0.0])

# Run NITER number of iterations:
for i in range(NITER):
	sim.reset()
	print "Running iteration", i
    # Inject all molecules into the central tet:
	sim.setTetCount(ctetidx, 'A', NINJECT)
	for j in range(ntpnts):
		sim.run(tpnts[j])
        # Loop over the tetrahedrons we are saving data for
		for k in range(SAMPLE):
            # Save the concentration in the tetrahedron, in uM
			res[i, j, k] = sim.getTetConc(tetidxs[k], 'A')*1.0e6
	printtime(time.time())

res_mean = numpy.mean(res, axis = 0)

########################################################################

def plotres(tidx):
	if (tidx >= INT/DT):
		print "Time index is out of range."
		return
	
	pylab.scatter(tetrads, res_mean[tidx], s=2)
	pylab.xlabel('Radial distance of tetrahedron ($\mu$m)')			
	pylab.ylabel('Concentration in tetrahedron ($\mu$M)')
	t = tpnts[tidx]
	pylab.title('Unbounded diffusion. Time: ' + str(t) + 's')
	plotanlyt(t)
	pylab.xlim(0.0, 10.0)
	pylab.ylim(0.0)
	pylab.show()

########################################################################

def plotanlyt(t): 	
	segs = 100 	
	anlytconc = numpy.zeros((segs)) 	
	radialds = numpy.zeros((segs)) 	
	maxrad = 0.0 	
	for i in tetrads: 		
		if (i > maxrad): maxrad = i 	
	maxrad *= 1e-6 	
	intervals = maxrad/segs 	
	rad = 0.0 	
	for i in range((segs)): 		
		# Find the conc from analytical solution, and convert to mol/L 		
		anlytconc[i]=1.0e3*(1/6.022e23)* \
            ((NINJECT/(math.pow((4*math.pi*DCST*t),1.5)))* \
                (math.exp((-1.0*(rad*rad))/(4*DCST*t)))) 		
		radialds[i] = rad*1e6 		
		rad += intervals 	
	pylab.plot(radialds, anlytconc, color = 'red')
	
########################################################################

plotres(100)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import datetime
import steps.model as smodel
import math
import numpy
import steps.solver as solvmod
import steps.geom as stetmesh
import steps.rng as srng
import steps.utilities.meshio as meshio
#from pylab import *
import time
import pylab

from tol_funcs import *

rng = srng.create('mt19937', 512) 
rng.initialize(int(time.time()%4294967295)) # The max unsigned long

NITER = 10
DT = 0.01
INT = 0.05

CONC =  50.0e-6	# the number of initial molecules

DCST = 0.1e-9

SAMPLE = 12033	

MESHFILE = 'cyl_diam0_4__len10_12Ktets'

# <1% fail with a tolerance of 5%
tolerance = 5.0/100


# create the array of tet indices to be found at random
tetidxs = numpy.zeros(SAMPLE, dtype = 'int')
# further create the array of tet barycentre distance to centre
tetrads = numpy.zeros(SAMPLE)

beg_time = time.time()

########################################################################
####                     ERROR FUNCTION STUFF                       ####
########################################################################

# Arguments are the x value to find, and the number of bins to iterate over
def erfunc(x, num = 1000):
	erf = 0.0
	
	place = 0.0
	ds= x/num
	for i in xrange(num):
		nowx = (i*x)/num 
		nextx = ((i+1)*x)/num
		goodx = (nowx+nextx)/2.0
		erf+=(ds*math.exp(-(goodx*goodx)))
	
	return 1 -(2*(erf/math.sqrt(math.pi)))
	
	
def getConc(Cs, D, x, t):
	return (Cs*erfunc(x/(math.sqrt(4*D*t))))
	
########################################################################

def gen_model():
   
	mdl = smodel.Model()
	X = smodel.Spec('X', mdl)
	cytosolv = smodel.Volsys('cytosolv', mdl)
	dif_X = smodel.Diff('diffX', cytosolv, X, dcst = DCST)
	
	return mdl

########################################################################

def gen_geom():
    mesh = meshio.loadMesh('validation/meshes/' +MESHFILE)[0]
    
    ntets = mesh.countTets()
    
    comp = stetmesh.TmComp('cyto', mesh, xrange(ntets))
    comp.addVolsys('cytosolv')
    
    # Now fill the array holding the tet indices to sample at random
    assert(SAMPLE == ntets)
    
    numfilled = 0
    while (numfilled < SAMPLE):
        tetidxs[numfilled] = numfilled
        numfilled +=1

    # Now find the distance of the centre of the tets to the Z lower face
    for i in range(SAMPLE):
        baryc = mesh.getTetBarycenter(tetidxs[i])
        min = mesh.getBoundMin()
        r = baryc[2] - min[2]
        # Convert to microns
        tetrads[i] = r*1.0e6
    
    return mesh

########################################################################

m = gen_model()
g = gen_geom()

# And fetch the total number of tets to make the data structures
ntets = g.countTets()

sim = solvmod.Tetexact(m, g, rng)

tpnts = numpy.arange(0.0, INT, DT)
ntpnts = tpnts.shape[0]


#Create the big old data structure: iterations x time points x concentrations
res = numpy.zeros((NITER, ntpnts, SAMPLE))

# Find the tets connected to the bottom face
# First find all the tets with ONE face on a boundary
boundtets = []

# store the 0to3 index of the surface triangle for each of these boundary tets
bt_srftriidx = []

for i in xrange(ntets):
	tettemp = g.getTetTetNeighb(i)
	if (tettemp[0] ==-1 or tettemp[1] == -1 or tettemp[2] == -1 or tettemp[3] == -1): 
		boundtets.append(i)
		templist = []
		if (tettemp[0] == -1): 
			templist.append(0)
		if (tettemp[1] == -1): 
			templist.append(1)
		if (tettemp[2] == -1): 
			templist.append(2)
		if (tettemp[3] == -1): 
			templist.append(3)
		bt_srftriidx.append(templist)

assert (boundtets.__len__() == bt_srftriidx.__len__())
		
minztets = []
boundminz = g.getBoundMin()[2] + 0.01e-06
num2s=0
for i in xrange(boundtets.__len__()):
	# get the boundary triangle
	if (bt_srftriidx[i].__len__() == 2): num2s+=1
	for btriidx in bt_srftriidx[i]:
		zminboundtri = True
		tribidx = g.getTetTriNeighb(boundtets[i])[btriidx]
		tritemp = g.getTri(tribidx)
		trizs = [0.0, 0.0, 0.0]
		trizs[0] = g.getVertex(tritemp[0])[2]
		trizs[1] = g.getVertex(tritemp[1])[2]
		trizs[2] = g.getVertex(tritemp[2])[2]
		for j in range(3):
			if (trizs[j]>boundminz): zminboundtri = False
		if (zminboundtri): minztets.append(boundtets[i])

nztets = minztets.__len__()
volztets = 0.0
for z in minztets:
	volztets += g.getTetVol(z)

for j in xrange(NITER):
    sim.reset()
    totset = 0
    for k in minztets:
        sim.setTetConc(k, 'X', CONC)
        sim.setTetClamped(k, 'X', True)
        totset+=sim.getTetCount(k, 'X')    
    for i in range(ntpnts):
        sim.run(tpnts[i])
        for k in range(SAMPLE):
            res[j, i, k] = sim.getTetCount(tetidxs[k], 'X')
    #print '%d / %d' % (j + 1, NITER)

itermeans = numpy.mean(res, axis = 0)
	
import pylab

########################################################################

tpnt_compare = [3, 4]
passed = True
max_err = 0.0

for t in tpnt_compare:
    NBINS=10
    radmax = 0.0
    radmin = 11.0
    for r in tetrads:
        if (r > radmax): radmax = r
        if (r < radmin) : radmin = r

    rsec = (radmax-radmin)/NBINS
    binmins = numpy.zeros(NBINS+1)
    tetradsbinned = numpy.zeros(NBINS)
    r = radmin
    bin_vols = numpy.zeros(NBINS)

    for b in range(NBINS+1):
        binmins[b] = r
        if (b!=NBINS): tetradsbinned[b] = r +rsec/2.0
        r+=rsec
    
    bin_counts = [None]*NBINS
    for i in xrange(NBINS):
        bin_counts[i] = []
    filled = 0

    for i in range(itermeans[t].size):
        irad = tetrads[i]
        
        for b in xrange(NBINS):
            if(irad>=binmins[b] and irad<binmins[b+1]):
                bin_counts[b].append(itermeans[t][i])
                bin_vols[b]+=sim.getTetVol(tetidxs[i])
                filled+=1.0
                break
    bin_concs = numpy.zeros(NBINS)
    for c in range(NBINS): 
        for d in range(bin_counts[c].__len__()):
            bin_concs[c] += bin_counts[c][d]
        bin_concs[c]/=(bin_vols[c])
        bin_concs[c]*=(1.0e-3/6.022e23)*1.0e6
    
    for i in range(NBINS):
        if (tetradsbinned[i] > 1 and tetradsbinned[i] < 4):
            rad = tetradsbinned[i]*1.0e-6
            det_conc =   (getConc(CONC*6.022e26, DCST, rad, tpnts[t])/6.022e26)*1.0e6         
            steps_conc = bin_concs[i]
            if not tolerable(det_conc, steps_conc, tolerance): passed = False
            if (abs(2*(det_conc-steps_conc)/(det_conc+steps_conc)) > max_err): max_err = abs(2*(det_conc-steps_conc)/(det_conc+steps_conc))
            #print "Error:", abs(2*(det_conc-steps_conc)/(det_conc+steps_conc))*100, "%"
    #print "Max error:", max_err*100, "%"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#  Last Changed Rev:  $Rev: 352 $
#  Last Changed Date: $Date: 2010-08-09 10:49:54 +0900 (Mon, 09 Aug 2010) $
#  Last Changed By:   $Author: wchen $

""" 
.. Note::

    This module is preliminary, means some of the functions are still under development.
    Code modification / debugging is wellcomed.
    Please email steps.dev@gmail.com if you would like to share you changes with others.

Mesh Control Utilities 

The meshctrl module provides functions for controling and manipulating the mesh, 
e.g. finding overlap surface triangles in meshes.

"""

import steps.geom
from math import *

################################################################################
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
################################################################################
def findOverlapTris(mesh, tets1, tets2):
    """
    Find overlap triangles between two sets of tetrahedrons within a mesh.
    
    Arguements:
        * steps.geom.Tetmesh mesh
        * list<uint> tets1
        * list<uint> tets2
        
    Return:
        list<uint>
    """
    
    tris1 = set()
    for i in tets1:
        tritemp = mesh.getTetTriNeighb(i)
        for j in range(4): 
            tris1.add(tritemp[j])
    tris2 = set()
    for i in tets2:
        tritemp = mesh.getTetTriNeighb(i)
        for j in range(4): 
            tris2.add(tritemp[j])

    common_tris = tris1.intersection(tris2)
    return list(common_tris)
    
################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

################################################################################

def findOverlapSurfTris(mesh1, mesh2):
    """
    Find overlap surface triangles between two meshes.
    Return a list of coupling data list formatted as
    
    Arguements:
        * steps.geom.Tetmesh mesh1
        * steps.geom.Tetmesh mesh2
        
    Return:
        list<list<uint, uint, uint, uint, float>>
    """
    
    surfs1 = mesh1.getSurfTris()
    surfs2 = mesh2.getSurfTris()
    data = []
    
    for s1 in surfs1:
        c1 = mesh1.getTriBarycenter(s1)
        
        neighb_tets1 = mesh1.getTriTetNeighb(s1)
        tetid1 = neighb_tets1[0]
        if tetid1 < 0:
            tetid1 = neighb_tets1[1]
        assert(tetid1 >= 0)
        tc1 = mesh1.getTetBarycenter(tetid1)
        
        for s2 in surfs2:
            c2 = mesh2.getTriBarycenter(s2)
            if (c1[0] == c2[0]) and (c1[1] == c2[1]) and (c1[2] == c2[2]): 
                neighb_tets2 = mesh2.getTriTetNeighb(s2)
                tetid2 = neighb_tets2[0]
                if tetid2 < 0:
                    tetid2 = neighb_tets2[1]
                assert(tetid2 >= 0)
                tc2 = mesh2.getTetBarycenter(tetid2)
                
                xdist = tc1[0] - tc2[0]
                ydist = tc1[1] - tc2[1]
                zdist = tc1[2] - tc2[2]
                dist = sqrt((xdist*xdist) + (ydist*ydist) + (zdist*zdist))
                
                coupling = [tetid1, s1, tetid2, s2, dist]
                data.append(coupling)
    return data
                
################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

################################################################################

def findSurfTrisInComp(mesh, comp_id):
    """
    Find surface triangles within a compartment.
    Return a list of surface triangle indices.
    
    Arguements:
        * steps.geom.Tetmesh mesh
        * string comp_id
        
    Return:
        list<uint>
    """
    
    surf_tris = mesh.getSurfTris()
    comp = mesh.getComp(comp_id)
    comp_tets = comp.getAllTetIndices()
    
    out_surfs = []
    for tet in  comp_tets:
        nb_tris = mesh.getTriTetNeighb(tet)
        for tri in nb_tris:
            if tri in surf_tris:
                out_surfs.append(tri)
    
    return out_surfs
            
################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

################################################################################

def findSurfTrisInTets(mesh, tet_list):
    """
    Find surface triangles within a compartment.
    Return a list of surface triangle indices.
    
    Arguements:
        * steps.geom.Tetmesh mesh
        * string comp_id
        
    Return:
        list<uint>
    """
    
    surf_tris = mesh.getSurfTris()

    out_surfs = []
    for tet in  tet_list:
        nb_tris = mesh.getTriTetNeighb(tet)
        for tri in nb_tris:
            if tri in surf_tris:
                out_surfs.append(tri)
    
    return out_surfs
            
################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

################################################################################


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Example: Diffusion boundary
# http://steps.sourceforge.net/manual/diffusion_boundary.html

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import steps.model as smodel
import steps.geom as sgeom
import steps.rng as srng
import steps.solver as solvmod
import steps.utilities.meshio as meshio

import numpy
from pylab import *
	
########################################################################

def gen_model():
    
    # Create the model container object
    mdl = smodel.Model()
    
    # Create the chemical species
    X = smodel.Spec('X', mdl)
    Y = smodel.Spec('Y', mdl)
    
    # Create separate volume systems for compartments A and B
    vsysA = smodel.Volsys('vsysA', mdl)
    vsysB = smodel.Volsys('vsysB', mdl)
    
    # Describe diffusion of molecules in compartments A and B
    # NOTE: diffusion is not defined for species X in compartment B
    diff_X_A = smodel.Diff('diff_X_A', vsysA, X, dcst = 0.1e-9)
    diff_X_B = smodel.Diff('diff_X_B', vsysB, X, dcst = 0.1e-9)
    diff_Y_A = smodel.Diff('diff_Y_A', vsysA, Y, dcst = 0.1e-9)    
    diff_Y_B = smodel.Diff('diff_Y_B', vsysB, Y, dcst = 0.1e-9) 
    
    # Return the container object
    return mdl
	
########################################################################

def gen_geom():
    mesh = meshio.loadMesh('meshes/cyl_len10_diam1')[0]
    ntets = mesh.countTets()
    
    tets_compA = []
    tets_compB = []
    tris_compA = set()
    tris_compB = set()
    z_max = mesh.getBoundMax()[2]
    z_min = mesh.getBoundMin()[2]
    z_mid = z_min+(z_max-z_min)/2.0
    
    for t in range(ntets):
        # Fetch the z coordinate of the barycenter
        barycz = mesh.getTetBarycenter(t)[2]
        # Fetch the triangle indices of the tetrahedron, a tuple of length 4:
        tris = mesh.getTetTriNeighb(t)
        if barycz < z_mid: 
            tets_compA.append(t)
            tris_compA.add(tris[0])
            tris_compA.add(tris[1])
            tris_compA.add(tris[2])
            tris_compA.add(tris[3])
        else: 
            tets_compB.append(t)
            tris_compB.add(tris[0])
            tris_compB.add(tris[1])
            tris_compB.add(tris[2])
            tris_compB.add(tris[3])
    
    # Create the mesh compartments
    compA = sgeom.TmComp('compA', mesh, tets_compA)
    compB = sgeom.TmComp('compB', mesh, tets_compB)
    compA.addVolsys('vsysA')
    compB.addVolsys('vsysB')
    
    # Make the diff boundary tris with the intersection and convert to list
    tris_DB = tris_compA.intersection(tris_compB)
    tris_DB = list(tris_DB)

    # Create the diffusion boundary between compA and compB
    diffb = sgeom.DiffBoundary('diffb', mesh, tris_DB)
    
    return mesh, tets_compA, tets_compB

########################################################################

mdl = gen_model()
mesh, tets_compA, tets_compB = gen_geom()

rng = srng.create_mt19937(256) 
rng.initialize(432) 

sim = solvmod.Tetexact(mdl, mesh, rng)
sim.reset()

tpnts = numpy.arange(0.0, 0.101, 0.001)
ntpnts = tpnts.shape[0]

# And fetch the total number of tets to make the data structures
ntets = mesh.countTets()
    
# Create the data structures: time points x tetrahedrons
resX = numpy.zeros((ntpnts, ntets))
resY = numpy.zeros((ntpnts, ntets))

tetx = mesh.findTetByPoint([0, 0, -4.99e-6])
tety = mesh.findTetByPoint([0,0, 4.99e-6])

sim.setTetCount(tetx , 'X', 1000)
sim.setTetCount(tety, 'Y', 500)

sim.setDiffBoundaryDiffusionActive('diffb', 'Y', True)

for i in range(ntpnts):
    sim.run(tpnts[i])
    for k in range(ntets):
        resY[i, k] = sim.getTetCount(k, 'Y')
        resX[i, k] = sim.getTetCount(k, 'X')

########################################################################

def plot_binned(t_idx, bin_n = 100):
    if (t_idx > tpnts.size):
        print "Time index is out of range."
        return
    
    z_tets = numpy.zeros(ntets)  
    zbound_min = mesh.getBoundMin()[2]
      
    # Now find the distance of the centre of the tets to the Z lower face
    for i in range(ntets):
        baryc = mesh.getTetBarycenter(i)
        z = baryc[2] - zbound_min
        # Convert to microns
        z_tets[i] = z*1.0e6
    
    # Find the maximum and minimum z of all tetrahedrons 
    z_max = z_tets.max()
    z_min = z_tets.min()
    
    # Set up the bin structures, recording the individual bin volumes
    z_seg = (z_max-z_min)/bin_n
    bin_mins = numpy.zeros(bin_n+1)
    z_tets_binned = numpy.zeros(bin_n)
    bin_vols = numpy.zeros(bin_n)    
    
    # Now sort the counts into bins for species 'X' 
    z = z_min
    for b in range(bin_n + 1):
        bin_mins[b] = z
        if (b!=bin_n): z_tets_binned[b] = z +z_seg/2.0
        z+=z_seg
    bin_counts = [None]*bin_n
    for i in range(bin_n): bin_counts[i] = []
    for i in range((resX[t_idx].size)):
        i_z = z_tets[i]
        for b in range(bin_n):
            if(i_z>=bin_mins[b] and i_z<bin_mins[b+1]):
                bin_counts[b].append(resX[t_idx][i])
                bin_vols[b]+=sim.getTetVol(i)
                break
    
    # Convert to concentration in arbitrary units              
    bin_concs = numpy.zeros(bin_n)
    for c in range(bin_n): 
        for d in range(bin_counts[c].__len__()):
            bin_concs[c] += bin_counts[c][d]
        bin_concs[c]/=(bin_vols[c]*1.0e18)
    
    t = tpnts[t_idx]
    
    scatter(z_tets_binned, bin_concs, label = 'X', color = 'blue')
    
    # Repeat the process for species 'Y'- separate from 'X' for clarity:
    z = z_min
    for b in range(bin_n + 1):
        bin_mins[b] = z
        if (b!=bin_n): z_tets_binned[b] = z +z_seg/2.0
        z+=z_seg
    bin_counts = [None]*bin_n
    for i in range(bin_n): bin_counts[i] = []
    for i in range((resY[t_idx].size)):
        i_z = z_tets[i]
        for b in range(bin_n):
            if(i_z>=bin_mins[b] and i_z<bin_mins[b+1]):
                bin_counts[b].append(resY[t_idx][i])
                break
                
    bin_concs = numpy.zeros(bin_n)
    for c in range(bin_n): 
        for d in range(bin_counts[c].__len__()):
            bin_concs[c] += bin_counts[c][d]
        bin_concs[c]/=(bin_vols[c]*1.0e18)
    
    scatter(z_tets_binned, bin_concs, label = 'Y', color = 'red')
    xlabel('Z axis (microns)', fontsize=16)
    ylabel('Bin concentration (N/m^3)', fontsize=16)
    ylim(0)    
    xlim(0, 10)
    legend(numpoints=1)
    show()

########################################################################

plot_binned(100, 50)




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Example: Surface-Volume reaction system (IP3 model)
# http://steps.sourceforge.net/manual/ip3.html

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import steps.model as smodel

#########################
# Model setup
#########################

mdl = smodel.Model()

# Calcium
Ca = smodel.Spec('Ca', mdl)

# IP3
IP3 = smodel.Spec('IP3', mdl)

############### receptor state objects ###############
# receptor state: 'naive' state (no bound ligands)
R = smodel.Spec('R', mdl)

# receptor state: bound IP3
RIP3 = smodel.Spec('RIP3', mdl)

# receptor state: bound IP3 and Ca (open)
Ropen = smodel.Spec('Ropen', mdl)

# receptor state: Ca bound to one inactivation site
RCa = smodel.Spec('RCa', mdl)

# receptor state: Ca bound to two inactivation sites
R2Ca = smodel.Spec('R2Ca', mdl)

# receptor state: Ca bound to three inactivation sites
R3Ca = smodel.Spec('R3Ca', mdl)

# receptor state: Ca bound to four inactivation sites
R4Ca = smodel.Spec('R4Ca', mdl)
#######################################################

# Surface system
surfsys = smodel.Surfsys('ssys', mdl)

# The 'forward' binding reactions:

R_bind_IP3_f = smodel.SReac('R_bind_IP3_f', surfsys, olhs=[IP3], slhs=[R], srhs=[RIP3])

RIP3_bind_Ca_f = smodel.SReac('RIP3_bind_Ca_f', surfsys, olhs=[Ca], slhs=[RIP3], srhs = [Ropen])

R_bind_Ca_f = smodel.SReac('R_bind_Ca_f', surfsys, olhs=[Ca], slhs=[R], srhs=[RCa])

RCa_bind_Ca_f = smodel.SReac('RCa_bind_Ca_f', surfsys, olhs=[Ca], slhs=[RCa],srhs = [R2Ca])

R2Ca_bind_Ca_f = smodel.SReac('R2Ca_bind_Ca_f', surfsys, olhs=[Ca], slhs= [R2Ca], srhs = [R3Ca])

R3Ca_bind_Ca_f = smodel.SReac('R3Ca_bind_ca_f', surfsys, olhs=[Ca], slhs=[R3Ca], srhs=[R4Ca])

# The 'backward' unbinding reactions:

R_bind_IP3_b = smodel.SReac('R_bind_IP3_b', surfsys, slhs=[RIP3], orhs=[IP3], srhs=[R])

RIP3_bind_Ca_b = smodel.SReac('RIP3_bind_Ca_b', surfsys, slhs=[Ropen], orhs=[Ca], srhs=[RIP3])

R_bind_Ca_b = smodel.SReac('R_bind_Ca_b', surfsys, slhs=[RCa], orhs=[Ca], srhs=[R])

RCa_bind_Ca_b = smodel.SReac('RCa_bind_Ca_b', surfsys, slhs=[R2Ca], orhs=[Ca], srhs=[RCa])

R2Ca_bind_Ca_b = smodel.SReac('R2Ca_bind_Ca_b', surfsys, slhs=[R3Ca], orhs=[Ca], srhs= [R2Ca])

R3Ca_bind_Ca_b = smodel.SReac('R3Ca_bind_ca_b', surfsys, slhs=[R4Ca], orhs=[Ca], srhs=[R3Ca])

# Ca ions passing through open IP3R channel
R_Ca_channel_f = smodel.SReac('R_Ca_channel_f', surfsys, ilhs=[Ca], slhs=[Ropen], orhs=[Ca], srhs=[Ropen])

R_bind_IP3_f.setKcst(1000e6)
R_bind_IP3_b.setKcst(25800)
RIP3_bind_Ca_f.setKcst(8000e6)
RIP3_bind_Ca_b.setKcst(2000)
R_bind_Ca_f.setKcst(8.889e6)
R_bind_Ca_b.setKcst(5)
RCa_bind_Ca_f.setKcst(20e6)
RCa_bind_Ca_b.setKcst(10)
R2Ca_bind_Ca_f.setKcst(40e6)
R2Ca_bind_Ca_b.setKcst(15)
R3Ca_bind_Ca_f.setKcst(60e6)
R3Ca_bind_Ca_b.setKcst(20)
R_Ca_channel_f.setKcst(2e8)

#########################
# Geom setup
#########################

import steps.geom as swm
wmgeom = swm.Geom()

# Create the cytosol compartment
cyt = swm.Comp('cyt', wmgeom)
cyt.setVol(1.6572e-19)

# Create the Endoplasmic Reticulum compartment
ER = swm.Comp('ER', wmgeom, vol = 1.968e-20)

# ER is the 'inner' compartment, cyt is the 'outer' compartment
memb = swm.Patch('memb', wmgeom, ER, cyt)
memb.addSurfsys('ssys')
memb.setArea(0.4143e-12)

print 'Inner compartment to memb is', memb.getIComp().getID()

print 'Outer compartment to memb is', memb.getOComp().getID()

#########################
# RNG setup
#########################
import steps.rng as srng
r = srng.create('mt19937', 512)
r.initialize(7233)

#########################
# Solver setup
#########################
import steps.solver as ssolver
sim = ssolver.Wmdirect(mdl, wmgeom, r)

NITER = 100
import pylab
import numpy
tpnt = numpy.arange(0.0, 0.201, 0.001)
res = numpy.zeros([NITER, 201, 2])
res_std = numpy.zeros([201, 2])
res_std1 = numpy.zeros([201, 2])
res_std2 = numpy.zeros([201, 2])


#########################
# Run simulation
#########################

for i in range (0, NITER):
    sim.reset()
    sim.setCompConc('cyt', 'Ca', 3.30657e-8)
    sim.setCompCount('cyt', 'IP3', 6)
    sim.setCompConc('ER', 'Ca', 150e-6)
    sim.setCompClamped('ER', 'Ca', True)
    sim.setPatchCount('memb', 'R', 160)
    for t in range(0,201):
        sim.run(tpnt[t])
        res[i,t,0] = sim.getPatchCount('memb', 'Ropen')
        res[i,t,1] = sim.getCompConc('cyt', 'Ca')
    pylab.plot(tpnt, res[i,:,0], color = 'blue', linewidth = 0.1)

res_mean = numpy.mean(res, 0)
res_std = numpy.std(res, 0)
res_std1 = res_mean[:,0] + res_std[:,0]
res_std2 = res_mean[:,0]- res_std[:,0]

pylab.plot(tpnt, res_mean[:,0], color = 'black', linewidth = 2.0, label = 'mean')
pylab.plot(tpnt, res_std1, color = 'gray', linewidth = 1.0, label='std')
pylab.plot(tpnt, res_std2,color = 'gray', linewidth = 1.0)

pylab.xlabel('Time (sec)')
pylab.ylabel('# IP3 receptors in open state')
pylab.title('IP3 receptor model: %d iterations with Wmdirect'%NITER)
pylab.ylim(0)
pylab.legend()
pylab.show()
        
        
        
        


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import datetime
import steps.model as smodel
import math
import steps.solver as solvmod
import steps.utilities.meshio as meshio
import steps.geom as stetmesh
import steps.rng as srng
import time
import numpy
import pylab

from tol_funcs import *

rng = srng.create('mt19937', 512) 
rng.initialize(int(time.time()%4294967295)) # The max unsigned long

NITER = 10
DT = 0.01
INT = 0.11

# The number of initial molecules:
NINJECT = 10000	

DCST = 0.2e-9

# In tests, with good code, <1% fail with a tolerance of 5%
tolerance = 5.0/100

# The number of tets to sample at random:
SAMPLE = 1060	

MESHFILE = 'cyl_diam2__len10_1060tets'

# create the array of tet indices to be found at random
tetidxs = numpy.zeros(SAMPLE, dtype = 'int')
# further create the array of tet barycentre distance to centre
tetrads = numpy.zeros(SAMPLE)

	
########################################################################

def gen_model():
    
    mdl = smodel.Model()
    
    X = smodel.Spec('X', mdl)
    cytosolv = smodel.Volsys('cytosolv', mdl)
    dif_X = smodel.Diff('diffX', cytosolv, X, dcst = DCST)
    
    return mdl
	
########################################################################

def gen_geom():
    mesh = meshio.loadMesh('validation/meshes/' +MESHFILE)[0]
    
    a = mesh.getBoundMax()[2]-mesh.getBoundMin()[2]
    area = mesh.getMeshVolume()/a
    
    ntets = mesh.countTets()
    comp = stetmesh.TmComp('cyto', mesh, range(ntets))
    comp.addVolsys('cytosolv')
    
    assert(SAMPLE == ntets)
    
    numfilled = 0
    while (numfilled < SAMPLE):
        tetidxs[numfilled] = numfilled
        numfilled +=1

    # Now find the distance of the centre of the tets to the Z lower face
    for i in range(SAMPLE):
        baryc = mesh.getTetBarycenter(tetidxs[i])
        min = mesh.getBoundMin()
        r = baryc[2] - min[2]
        # Convert to microns
        tetrads[i] = r*1.0e6
    
    return mesh, area, a

########################################################################

m = gen_model()
g, area, a = gen_geom()

# And fetch the total number of tets to make the data structures
ntets = g.countTets()

sim = solvmod.Tetexact(m, g, rng)

tpnts = numpy.arange(0.0, INT, DT)
ntpnts = tpnts.shape[0]


# Create the big old data structure: iterations x time points x concentrations
res = numpy.zeros((NITER, ntpnts, SAMPLE))

# Find the tets connected to the bottom face
# First find all the tets with ONE face on a boundary
boundtets = []
#store the 0to3 index of the surface triangle for each of these boundary tets
bt_srftriidx = []

for i in xrange(ntets):
	tettemp = g.getTetTetNeighb(i)
	if (tettemp[0] ==-1 or tettemp[1] == -1 or tettemp[2] == -1 or tettemp[3] == -1): 
		boundtets.append(i)
		templist = []
		if (tettemp[0] == -1): 
			templist.append(0)
		if (tettemp[1] == -1): 
			templist.append(1)
		if (tettemp[2] == -1): 
			templist.append(2)
		if (tettemp[3] == -1): 
			templist.append(3)
		bt_srftriidx.append(templist)

assert (boundtets.__len__() == bt_srftriidx.__len__())
		
minztets = []
boundminz = g.getBoundMin()[2] + 0.01e-06
num2s=0
for i in xrange(boundtets.__len__()):
	# get the boundary triangle
	if (bt_srftriidx[i].__len__() == 2): num2s+=1
	for btriidx in bt_srftriidx[i]:
		zminboundtri = True
		tribidx = g.getTetTriNeighb(boundtets[i])[btriidx]
		tritemp = g.getTri(tribidx)
		trizs = [0.0, 0.0, 0.0]
		trizs[0] = g.getVertex(tritemp[0])[2]
		trizs[1] = g.getVertex(tritemp[1])[2]
		trizs[2] = g.getVertex(tritemp[2])[2]
		for j in range(3):
			if (trizs[j]>boundminz): zminboundtri = False
		if (zminboundtri): minztets.append(boundtets[i])

nztets = minztets.__len__()
volztets = 0.0
for z in minztets:
	volztets += g.getTetVol(z)
conc = NITER*6.022e23*1.0e-3/volztets

for j in xrange(NITER):
    sim.reset()
    tetcount = int((1.0*NINJECT)/nztets)
    totset = 0
    for k in minztets:
        sim.setTetCount(k, 'X', tetcount)
        totset+=tetcount
    for i in range(ntpnts):
        sim.run(tpnts[i])
        for k in range(SAMPLE):
            res[j, i, k] = sim.getTetCount(tetidxs[k], 'X')
    #print '%d / %d' % (j + 1, NITER)

itermeans = numpy.mean(res, axis = 0)
	

########################################################################

D = DCST
pi = math.pi
nmax = 1000
N = NINJECT
N = int((1.0*NINJECT)/nztets)*nztets
def getprob(x,t):
	if(x>a): 
		print 'x out of bounds'
		return
	p=0.0
	for n in xrange(nmax):
		if (n==0): A = math.sqrt(1.0/a)
		else : A = math.sqrt(2.0/a)
		p+= math.exp(-D*math.pow((n*pi/a), 2)*t)*A*math.cos(n*pi*x/a)*A*a
	
	return p*N/a

tpnt_compare = [6, 8, 10]
passed = True
max_err = 0.0

for t in tpnt_compare:
    NBINS = 5
    
    radmax = 0.0
    radmin = 11.0
    for r in tetrads:
        if (r > radmax): radmax = r
        if (r < radmin) : radmin = r

    rsec = (radmax-radmin)/NBINS
    binmins = numpy.zeros(NBINS+1)
    tetradsbinned = numpy.zeros(NBINS)
    r = radmin
    bin_vols = numpy.zeros(NBINS)

    for b in range(NBINS+1):
        binmins[b] = r
        if (b!=NBINS): tetradsbinned[b] = r +rsec/2.0
        r+=rsec
    
    bin_counts = [None]*NBINS
    for i in xrange(NBINS):
        bin_counts[i] = []
    filled = 0

    for i in range(itermeans[t].size):
        irad = tetrads[i]
        
        for b in xrange(NBINS):
            if(irad>=binmins[b] and irad<binmins[b+1]):
                bin_counts[b].append(itermeans[t][i])
                bin_vols[b]+=sim.getTetVol(tetidxs[i])
                filled+=1.0
                break
    bin_concs = numpy.zeros(NBINS)
    for c in range(NBINS): 
        for d in range(bin_counts[c].__len__()):
            bin_concs[c] += bin_counts[c][d]
        bin_concs[c]/=(bin_vols[c])
        bin_concs[c]*=(1.0e-3/6.022e23)*1.0e6
    
    for i in range(NBINS):
        if (tetradsbinned[i] > 2 and tetradsbinned[i] < 8):
            rad = tetradsbinned[i]*1.0e-6
            det_conc = (getprob(rad, tpnts[t])/area)*(1.0/6.022e20)
            steps_conc = bin_concs[i]
            if not tolerable(det_conc, steps_conc, tolerance): passed = False
            if (abs(2*(det_conc-steps_conc)/(det_conc+steps_conc)) > max_err): max_err = abs(2*(det_conc-steps_conc)/(det_conc+steps_conc))
            #print "Error:",abs(2*(det_conc-steps_conc)/(det_conc+steps_conc))*100.0, "%"        
#print "max error", max_err*100.0, "%"               





# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import steps.model as smodel
import steps.geom as swm
import steps.rng as srng
import steps.solver as ssolver

##############################
# Model Setup
##############################

mdl = smodel.Model()

molA = smodel.Spec('molA', mdl)
molB = smodel.Spec('molB', mdl)
molC = smodel.Spec('molC', mdl)

vsys = smodel.Volsys('vsys', mdl)

kreac_f = smodel.Reac('kreac_f', vsys, lhs=[molA,molB], rhs=[molC], kcst=0.3e6)
kreac_b = smodel.Reac('kreac_b', vsys, lhs=[molC], rhs=[molA,molB])
kreac_b.kcst = 0.7

##############################
# Geom Setup
##############################
wmgeom = swm.Geom()

comp = swm.Comp('comp', wmgeom)
comp.addVolsys('vsys')
comp.setVol(1.6667e-21)

##############################
# RNG Setup
##############################
r = srng.create('mt19937', 256)
r.initialize(23412)


##############################
# Well Mixed DirectSSA
##############################
sim = ssolver.Wmdirect(mdl, wmgeom, r)

sim.reset()

sim.setCompConc('comp', 'molA', 31.4e-6)
sim.setCompConc('comp', 'molB', 22.3e-6)

###################################
# Checkpoint manually
###################################
print "Manually checkpoint"
sim.checkpoint("manual.checkpoint")
print "Sim Time: %e" % sim.getTime()
print "Sim Steps: %e" % sim.getNSteps()

###################################
# Run Simulation with checkpointing
###################################
sim.run(1e-2, 1e-3)
print "Sim Time: %e" % sim.getTime()
print "Sim Steps: %e" % sim.getNSteps()

###################################
# Restore from checkpoint file
###################################
print "Restore from file"
sim.restore("manual.checkpoint")
print "Sim Time: %e" % sim.getTime()
print "Sim Steps: %e" % sim.getNSteps()

###################################
# RK4 solver
###################################
print "Well Mixed RK4"
rk4sim = ssolver.Wmrk4(mdl, wmgeom, r)

rk4sim.reset()
rk4sim.setDT(1e-5)
rk4sim.setCompConc('comp', 'molA', 31.4e-6)
rk4sim.setCompConc('comp', 'molB', 22.3e-6)

rk4sim.checkpoint("rk4.cp")
print "A count: ", rk4sim.getCompCount("comp", "molA")
print "B count: ", rk4sim.getCompCount("comp", "molB")
print "C count: ", rk4sim.getCompCount("comp", "molC")

rk4sim.advance(1e-2, 1e-3)
print "Sim Time: %e" % rk4sim.getTime()

print "A count: ", rk4sim.getCompCount("comp", "molA")
print "B count: ", rk4sim.getCompCount("comp", "molB")
print "C count: ", rk4sim.getCompCount("comp", "molC")

print "restore"
rk4sim.restore("rk4.cp")
print "A count: ", rk4sim.getCompCount("comp", "molA")
print "B count: ", rk4sim.getCompCount("comp", "molB")
print "C count: ", rk4sim.getCompCount("comp", "molC")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import datetime
import steps.model as smodel
import math
import numpy
import steps.solver as solvmod
import steps.utilities.meshio as smeshio
import steps.geom as stetmesh
import steps.rng as srng
import time
import pylab

from tol_funcs import *

rng = srng.create('mt19937', 1024) 
rng.initialize(int(time.time()%4294967295)) # The max unsigned long

# Number of iterations; plotting dt; sim endtime:
NITER = 10

DT = 0.01
INT = 0.21

# Number of molecules injected in centre; diff constant; number of tets sampled:
NINJECT = 100000 	

DCST = 0.02e-9

# With good code <1% fail with a tolerance of 5% 
tolerance = 5.0/100

########################################################################

SAMPLE = 32552	 # all tets

MESHFILE = 'sphere_rad10_33Ktets_adaptive'

# create the array of tet indices to be found at random
tetidxs = numpy.zeros(SAMPLE, dtype = 'int')
for i in range(SAMPLE): tetidxs[i] = i

# further create the array of tet barycentre distance to centre
tetrads = numpy.zeros(SAMPLE)
tetvols = numpy.zeros(SAMPLE)

########################################################################

def gen_model():
   
    mdl = smodel.Model()
    X = smodel.Spec('X', mdl)
    
    cytosolv = smodel.Volsys('cytosolv', mdl)
    dif_X = smodel.Diff('diffX', cytosolv, X, dcst = DCST)

    return mdl
	
########################################################################

def gen_geom():
    mesh = smeshio.loadMesh('validation/meshes/'+MESHFILE)[0]
    ctetidx = mesh.findTetByPoint([0.0, 0.0, 0.0])

    ntets = mesh.countTets()
    comp = stetmesh.TmComp('cyto', mesh, xrange(ntets))
    comp.addVolsys('cytosolv')
    
    # Now find the distance of the centre of the tets to the centre of the centre tet (at 0,0,0)
    cbaryc = mesh.getTetBarycenter(ctetidx)
    for i in range(SAMPLE):
        baryc = mesh.getTetBarycenter(tetidxs[i])
        r2 = math.pow((baryc[0]-cbaryc[0]),2) + math.pow((baryc[1]-cbaryc[1]),2) + math.pow((baryc[2]-cbaryc[2]),2)
        r = math.sqrt(r2)
        # Conver to microns
        tetrads[i] = r*1.0e6
        tetvols[i] = mesh.getTetVol(tetidxs[i])
    
    return mesh

########################################################################

m = gen_model()
g = gen_geom()

# Fetch the index of the centre tet
ctetidx = g.findTetByPoint([0.0, 0.0, 0.0])
# And fetch the total number of tets to make the data structures
ntets = g.countTets()

sim = solvmod.Tetexact(m, g, rng)

tpnts = numpy.arange(0.0, INT, DT)
ntpnts = tpnts.shape[0]

#Create the big old data structure: iterations x time points x concentrations
res = numpy.zeros((NITER, ntpnts, SAMPLE))

for j in xrange(NITER):
    sim.reset()
    sim.setTetCount(ctetidx, 'X', NINJECT)
    for i in range(ntpnts):
        sim.run(tpnts[i])
        for k in range(SAMPLE):
            res[j, i, k] = sim.getTetCount(tetidxs[k], 'X')
    #print '%d / %d' % (j + 1, NITER)

itermeans = numpy.mean(res, axis = 0)


tpnt_compare = [10, 15, 20]
passed = True
max_err = 0.0

for t in tpnt_compare:
    bin_n = 20
    
    r_max = tetrads.max()
    r_min = 0.0
    
    r_seg = (r_max-r_min)/bin_n
    bin_mins = numpy.zeros(bin_n+1)
    r_tets_binned = numpy.zeros(bin_n)
    bin_vols = numpy.zeros(bin_n)    
    
    r = r_min
    for b in range(bin_n + 1):
        bin_mins[b] = r
        if (b!=bin_n): r_tets_binned[b] = r +r_seg/2.0
        r+=r_seg
    bin_counts = [None]*bin_n
    for i in range(bin_n): bin_counts[i] = []
    for i in range((itermeans[t].size)):
        i_r = tetrads[i]
        for b in xrange(bin_n):
            if(i_r>=bin_mins[b] and i_r<bin_mins[b+1]):
                bin_counts[b].append(itermeans[t][i])
                bin_vols[b]+=sim.getTetVol(tetidxs[i])
                break
                
    bin_concs = numpy.zeros(bin_n)
    for c in range(bin_n): 
        for d in range(bin_counts[c].__len__()):
            bin_concs[c] += bin_counts[c][d]
        bin_concs[c]/=(bin_vols[c]*1.0e18)
    
    for i in range(bin_n):
        if (r_tets_binned[i] > 2.0 and r_tets_binned[i] < 6.0):
            rad = r_tets_binned[i]*1.0e-6
            det_conc = 1e-18*((NINJECT/(math.pow((4*math.pi*DCST*tpnts[t]),1.5)))*(math.exp((-1.0*(rad*rad))/(4*DCST*tpnts[t]))))
            steps_conc = bin_concs[i]
            if not tolerable(det_conc, steps_conc, tolerance): passed = False
"""            if (abs(2*(det_conc-steps_conc)/(det_conc+steps_conc)) > max_err): max_err = abs(2*(det_conc-steps_conc)/(det_conc+steps_conc))
    print max_err*100.0, "%"        
print "max error", max_err*100.0, "%"            
"""        




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import steps.model as smod
import steps.geom as sgeom
import steps.rng as srng
import steps.solver as ssolv

import math
import time
import steps.utilities.meshio as smeshio
import random
import steps.utilities.visual as svisual
########################################################################

DT = 0.001			# Sampling time-step
INT = 1		# Sim endtime

DCSTA = 400*1e-12
DCSTB = DCSTA
RCST = 100000.0e6

NA0 = 1000			# Initial number of A molecules
NB0 = NA0		# Initial number of B molecules

scale = 1.0e-6
OUTPUTFILE = "reac_diff.trace"
########################################################################
rng = srng.create('mt19937', 16384) 
rng.initialize(int(time.time()%4294967295)) # The max unsigned long


mdl  = smod.Model()

A = smod.Spec('A', mdl)
B = smod.Spec('B', mdl)
C = smod.Spec('C', mdl)

volsys = smod.Volsys('vsys',mdl)


R1 = smod.Reac('R1', volsys, lhs = [A,B], rhs = [C])

R1.setKcst(RCST)

D_a = smod.Diff('D_a', volsys, A, dcst = DCSTA)
D_b = smod.Diff('D_b', volsys, B, dcst = DCSTB)
D_c = smod.Diff('D_c', volsys, C, dcst = DCSTB)

mesh, nodeproxy, tetproxy, triproxy = smeshio.importAbaqus('brick_40_4_4_2_comp.inp', scale)
ntets = mesh.ntets
tet_groups = tetproxy.blocksToGroups()
Atets = tet_groups['EB1']
nATets = len(Atets)
Btets = tet_groups['EB2']
nBTets = len(Btets)
comp1 = sgeom.TmComp('comp1', mesh, xrange(mesh.ntets))
comp1.addVolsys('vsys')


print "create sim"
sim = ssolv.Tetexact(mdl, mesh, rng)

svisual.GUISim(mdl, mesh, mesh)
 




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#  Last Changed Rev:  $Rev: 410 $
#  Last Changed Date: $Date: 2011-04-07 16:11:28 +0900 (Thu, 07 Apr 2011) $
#  Last Changed By:   $Author: iain $

""" 
.. Note::

    This module is preliminary, means some of the functions are still under development.
    Code modification / debugging is wellcomed.
    Please email steps.dev@gmail.com if you would like to share you changes with others.

Mesh Import and Export Utilities 

The MeshIO utilities are designed for importing geometry data from outside mesh-generators,
as well as creating STEPS mesh object and exporting it using STEPS' own format.

Current importing support includes:

* Abaqus data format (http://www.simulia.com)

  Abaqus format is supported by CUBIT (http://cubit.sandia.gov/) 
  and Netgen (http://sourceforge.net/projects/netgen-mesher/)
  
* Tengen's data formats (http://tetgen.berlios.de/)

  Currently, only the .node, .ele and .face files are supported, 
  other Tetgen files is not used in STEPS.

Each import function (either importTetgen or importAbaqus) will read and import data 
from given data file(s), returns a Tetmesh object for STEPS simulation, and three 
ElementProxy objects where geometry data and id mapping are stored. 

Once a Tetmesh object has been created the data can be saved to two separate files 
by the saveTetmesh() function: 

* An xml annotated file containing the nodes, triangles and tetrahedra.

* A text file containing further information needed by STEPS solvers found 
  in Tetmesh object constructor. 

These files can later be loaded by the loadTetmesh() method and reconstruct
the Tetmesh object. This is intened to drastically reduce mesh-loading time
for large meshes (over ~100,000 voxels). By storing all data required by STEPS
internallyin these two files this infomation does not have to be found each time
by the Tetmesh object constructor.

**IMPORTANT NOTICE**

User is recommanded to save the tetmesh objects using the saveTetmesh() method
and recreate the objects from the saved files, instead of creating the objects
via the importing functions each time, if the tetmesh objects are intented to
be used in multiple simulations. Since the importing functions require a massive
amount of time to create the Tetmesh object, comparing to the loadTetmesh() method.

"""

import time
import steps.geom as stetmesh
import os.path as opath

#############################################################################################

class ElementProxy:
    """ Element Proxy Object

    The Element Proxy object contains data imported from outside files (Abaqus, Tetgen),
    as well as the mapping from the element's import id to its STEPS id, and vice versa.
    it also records block and/or group information contained in the imported file, which
    could be used for data tracing and compartment/patch creactions::

        ElementProxy(type, unitlength)
    
    Construct an empty ElementProxy object.
    
    Parameters:
    
    * type (string): The type of the elements stored in this object.
    
    * unitlength (int): The length of vector units required for each element data.
      e.g. A node with 3 coordinates requires unitlength of 3,
      a tetrahedron with 4 nodes requires unitlength of 4.
      A triangle requires unitlength of 3.
    
    Example::
    
        nodeproxy = ElementProxy('node', 3)
    
    """

    def __init__(self, type, unitlength):
        self.type = type
        self.unitlength = unitlength
        self.idcounter = 0
        self.data = []
        self.importid = []
        self.stepsid = {}
        self.blocks = {}
        self.groups = {}
        self.tempblockname = ''
        self.tempblockstart = -1

    def insert(self, import_id, import_data):
        """
        Insert an element to the Element Map object, a STEPS id will be assigned automatically.

        Parameters:
        
        * import_id (int): The id of the element given by the importing file.
        * import_data (list): a list of data belongs to the element, 
          e.g. for a node, a 3 units list of its coordinates.

        Example:
        
        nodeproxy.insert(1, [0.1, 0.2, 0.4])
        
        Insert a node with import id 1 and coordinates x = 0.1, y = 0.2, z = 0.4 to nodeproxy.
        type nodeproxy.getSTEPSID(1) will return the STEPS id of this element.

        """
        self.data += import_data
        self.importid.append(import_id)
        self.stepsid[import_id] = self.idcounter
        self.idcounter += 1
     
    def getType(self):
        """
        Return the type of elements stored in this Element Map object.

        """
        return self.type

    def getSize(self):
        """
        Return the number of elements stored in the Element Map object.

        """
        return self.idcounter
        
    def getDataFromSTEPSID(self, steps_id):
        """
        Return the data of element with STEPS id steps_id.

        Parameters:
        
        * steps_id (int): The STEPS id of the element, this id is automatically allocated when
          inserting the element to the Element Map (see insert() method).

        """              
        return self.data[steps_id * self.unitlength : (steps_id + 1) * self.unitlength]
    
    def getDataFromImportID(self, import_id):
        """
        Return the data of element with import id import_id.

        Parameters:
        
        import_id (int): 
        The import id of the element, this id is given by outside generator
        and contained in the imported file.

        """   
        return self.getDataFromSTEPSID(self.stepsid[import_id])

    def getAllData(self):
        """
        Return all data as an one dimentional list. 
        This list can be used to construct the Tetmesh.

        Example::
        
            nodedata = nodeproxy.getAllData()
            tetdata = tetproxy.getAllData()
            mesh = steps.geom.Tetmesh(nodedata, tetdata)

        """
        return self.data

    def getSTEPSID(self, import_id):
        """
        Return the STEPS id of a element from its import id.

        Parameters:
        
        * import_id (int): The import id of the element.

        """
        return self.stepsid[import_id]

    def getImportID(self, steps_id):
        """
        Return the import id of a element from its STEPS id.

        Parameters:
        
        * steps_id (int): The STEPS id of the element.

        """

        return self.importid[steps_id]

    def addGroup(self, group_name, group_ids):
        """
        Add a group of elements in the Element Map object.

        An group is defined as a list of element ids. Unlike blocks (see blockBegin() method),
        the ids of a group need not be continual. For this reason, user should provide a list
        which contains all ids of elements belong to this group. Each group has a unique name
        for data access.

        No uniqueness or order is predefined for a group, this can be defined by the user for
        specific usage.

        Groups can be used to construct compartments and patches in STEPS, please refer to the
        user manual for examples about using groups and blocks.
        
        Parameters:
        
        * group_name (string): Name of the element group.
        * group_ids (list): A list of ids of elements belong to the group.

        """

        self.groups[group_name] = group_ids
        
    def addToGroup(self, group_name, id):
        if (group_name in self.groups):
            self.groups[group_name].append(id)
        else:
            self.groups[group_name] = [id]

    def getGroups(self):
        """
        Return all groups stored in the Element Map object.

        Example:
        
        The following group dictionary contains two groups::
            
            "Group1" contains element ids 1,2,6,4,8
            "Group2" contains element ids 3,1,2,1

            groups = {"Group1":[1,2,6,4,8], "Group2":[3,1,2,1]}

        Note that element 1 appears twice in "Group2".

        """

        return self.groups
        
    def blockBegin(self, block_name):
        """
        Notify the Element Map object that a new block with name block_name begins from
        the next element.

        A block is a special type of group whose element's STEPS ids is continual,
        i.e. can be represented by a begin id and an end id. An unique name should be 
        given to each block, which can be used to access associated block information
        (i.e. a two unit list contains the block begin id and block end id).
        
        If another block was initialized before calling the blockEnd() method,
        this method will also end the previous block and the its data will be stored.

        A dictionary of blocks can be converted to a dictionary of group via the 
        blocksToGroups() method.

        Notice:
        
        Although a block will be finished and stored when a new block is notified without
        calling the blockEnd() method, user is recommanded to call blockEnd() in the usual
        case.

        Parameters:
        
        * block_name (string): Name of the element block.

        """

        if (self.tempblockstart != -1):
            self.blocks[self.tempblockname] = [self.tempblockstart, self.idcounter - 1]
        self.tempblockname = block_name
        self.tempblockstart = self.idcounter

    def blockEnd(self):
        """
        Notify the Element Map object that the current element block should be ended.
        Block information (i.e. a dictionary element blockname:[begin_id, end_id]) will
        be added to the ElementProxy object.

        Notice:
        
        If the blockEnd() method is called before any block begins, or a block has been
        ended and there is no new block started, no information will be added to the block
        dictionary.

        """
        if (self.tempblockstart != -1):
            self.blocks[self.tempblockname] = [self.tempblockstart, self.idcounter - 1]
            self.tempblockstart = -1

    def getBlocks(self):
        """
        Return the block dictionary stored in the Element Map object.

        A block dictionary uses the unique block names as keys, and a list of the start and
        end ids as values.

        Example:
        
        The following block dictionary contains two blocks: "Block1" starts from element with
        id of 0 and end at element 100, "Block2" starts from element 101 and ends at element 110::

            blocks = { "Block1":[0, 100], "Block2":[100, 110] }

        To access individual blocks, for example "Block1", type::
            
            block1 = blocks["Block1"]

        User is recommanded to check Python's manual for the usage of dictionary.

        """

        return self.blocks


    def blocksToGroups(self):
        """
        Convert the block dictionary to a group dictionary.

        Return:
        
        A dictionary of groups.

        """
        
        it = iter(self.blocks)
        converted_groups = {}
        for key in it:
            blockrange = self.blocks[key]
            converted_groups[key] = list(range(blockrange[0], blockrange[1] + 1))
        return converted_groups



#############################################################################################

def importTetGen(pathroot):
    """
    Read a TetGen-generated or refined mesh from a set of files.
    
    The TetGen file format for describing a mesh actually consists of 
    multiple files with varying suffixes. Currently, this routine only
    reads meshes consisting of three files:
    
    * <input>.node: describing the tetrahedral mesh node points.
    * <input>.ele: describing tetrahedral elements, each of which
      consists of 4 pointers into the <input>.node file. (TetGen
      also supports 10-node elements; these 6 extra nodes are obviously 
      not read by STEPS.) 
    * <input>.face: describing triangular faces, each of which 
      consists of 3 pointers into the <input>.node file. This file is optional.
    
    Other files are .vol (list of maximum volumes), .var (variant constraints)
    .neigh (list of neighbours), .smesh (simple PLC descriptions) and .edge
    (list of boundary edges) files. None of these seem relevant for our
    use cases, so we don't load them even when they are there. In particular,
    the .neigh file is computed by STEPS itself.
    
    Please refer to the TetGen manual (pages 31-40 in the last edition)
    for more information on these file formats
    
    tetgen.berlios.de/files/tetgen-manual.pdf
    
    (See the documentation for steps.geom.tetmesh for details about the mesh object.) 
    
    PARAMETERS:
    
    * pathroot
      The root of the path name. E.g. mesh/torus would make this
      routine try to read files mesh/torus.node, mesh/torus.ele 
      and optionally for mesh/torus.face
    
    RETURNS:
    
    mesh, nodeproxy, tetproxy, triproxy 

    mesh: The STEPS TetMesh object
    nodeproxy: Element Map for nodes
    tetproxy: Element Map for tetrahedrons
    triproxy: Element Map for triangles

   IMPORTANT NOTICE:
   
   User is recommanded to save the tetmesh objects using the saveTetmesh() method
   and recreate the objects from the saved files, instead of creating the objects
   via the importing functions each time, if the tetmesh objects are intented to
   be used in multiple simulations. Since the importing functions require a massive
   amount of time to create the Tetmesh object, comparing to the loadTetmesh() method.

    """
    nodefname = pathroot + '.node'
    elefname = pathroot + '.ele'
    facefname = pathroot + '.face'

    nodeproxy = ElementProxy('node', 3)
    tetproxy = ElementProxy('tet', 4)
    triproxy = ElementProxy('tri', 3) 

    # Is there a .node file?
    if not opath.isfile(nodefname):
        print(nodefname)
        return None
    if not opath.isfile(elefname):
        print(elefname)
        return None
    if not opath.isfile(facefname):
        facefname = ''
    
    # Try to read the node file.
    nodefile = open(nodefname, 'r')
    # First line is:  <x> <y> <z> [att<# of points> <dimension (3)> <# of attributes>
    #                <boundary marker (0 or 1)>
    line = nodefile.readline()
    while (line[0] == '#' or line[0] == '\n'):
        line = nodefile.readline()
    tokens = line.split()
    assert len(tokens) == 4
    nnodes = int(tokens[0])
    assert nnodes > 0
    ndims = int(tokens[1])
    assert ndims == 3
    nattribs = int(tokens[2])
    bmarkers = int(tokens[3])
    while (nodeproxy.getSize() != nnodes):
        line = nodefile.readline()
        commentstart = line.find('#')
        if commentstart != -1:
            line = line[0:commentstart]
        # Remaing lines: <point #>ributes] 
        #                [boundary marker]
        tokens = line.split()
        if len(tokens) == 0: 
            continue
        nodeid = int(tokens[0])
        node = [0.0, 0.0, 0.0]
        node[0] = float(tokens[1])
        node[1] = float(tokens[2])
        node[2] = float(tokens[3])
        nodeproxy.insert(nodeid, node)
    # Close the file.
    nodefile.close()

    # Try to read the tet file.
    elefile = open(elefname, 'r')
    line = elefile.readline()
    while (line[0] == '#' or line[0] == '\n'):
        line = elefile.readline()
    tokens = line.split()
    assert len(tokens) == 3
    ntets = int(tokens[0])
    assert ntets > 0
    ndims = int(tokens[1])
    assert ndims == 4
    nattribs = int(tokens[2])
    while (tetproxy.getSize() != ntets):
        line = elefile.readline()
        commentstart = line.find('#')
        if commentstart != -1:
            line = line[0:commentstart]
        # Remaing lines: <point #>ributes] 
        #                [boundary marker]
        tokens = line.split()
        if len(tokens) == 0: 
            continue
        tetid = int(tokens[0])
        tet = [0, 0, 0, 0]
        tet[0] = nodeproxy.getSTEPSID(int(tokens[1]))
        tet[1] = nodeproxy.getSTEPSID(int(tokens[2]))
        tet[2] = nodeproxy.getSTEPSID(int(tokens[3]))
        tet[3] = nodeproxy.getSTEPSID(int(tokens[4]))
        tetproxy.insert(tetid, tet)
        if nattribs == 1:
            tetproxy.addToGroup(tokens[5], tetproxy.getSTEPSID(tetid))
    elefile.close()

    # Try to read the tri file.
    if (facefname != ''):
        facefile = open(facefname, 'r')
        line = facefile.readline()
        while (line[0] == '#' or line[0] == '\n'):
            line = facefile.readline()
        tokens = line.split()
        assert len(tokens) == 2
        ntris = int(tokens[0])
        assert ntris > 0
        nattribs = int(tokens[1])
        while (triproxy.getSize() != ntris):
            line = facefile.readline()
            commentstart = line.find('#')
            if commentstart != -1:
                line = line[0:commentstart]
            # Remaing lines: <point #>ributes] 
            #                [boundary marker]
            tokens = line.split()
            if len(tokens) == 0: 
                continue
            triid = int(tokens[0])
            tri = [0, 0, 0]
            tri[0] = nodeproxy.getSTEPSID(int(tokens[1]))
            tri[1] = nodeproxy.getSTEPSID(int(tokens[2]))
            tri[2] = nodeproxy.getSTEPSID(int(tokens[3]))
            triproxy.insert(triid, tri)
            if nattribs == 1:
                triproxy.addToGroup(tokens[4], triproxy.getSTEPSID(triid))
    # Close the file.
    facefile.close()

    print("Read TetGen files succesfully")    
    
    nodedata = nodeproxy.getAllData()
    tetdata = tetproxy.getAllData() 
    tridata = triproxy.getAllData() 

    print("creating Tetmesh object in STEPS...")
    mesh = stetmesh.Tetmesh(nodedata, tetdata, tridata)
    print("Tetmesh object created.")
    return mesh, nodeproxy, tetproxy, triproxy
    

#############################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#############################################################################################

def importAbaqus(filename, scale, ebs = None):
    """ 
    Read a ABAQUS-formated mesh file, return the created steps.geom.Tetmesh object, 
    the element mapping for nodes, tetraedrons and triangles.
    
    PARAMETERS:
    
    * filename: the Abaqus filename (or path) including any suffix.
    * scale: LENGTH scale from the importing mesh to real geometry. e.g. a radius 
      of 10 in the importing file to a radius of 1 micron in STEPS, scale is 1e-7.
    * ebs: specify the names of selected element blocks which are included in the mesh
    
    RETURNS:
    
    mesh, nodeproxy, tetproxy, triproxy 

    * mesh: The STEPS TetMesh object
    * nodeproxy: Element Map for nodes
    * tetproxy: Element Map for tetrahedrons
    * triproxy: Element Map for triangles

   IMPORTANT NOTICE:
   
   User is recommanded to save the tetmesh objects using the saveTetmesh() method
   and recreate the objects from the saved files, instead of creating the objects
   via the importing functions each time, if the tetmesh objects are intented to
   be used in multiple simulations. Since the importing functions require a massive
   amount of time to create the Tetmesh object, comparing to the loadTetmesh() method.

    """
    print("Reading Abaqus file...")
    btime = time.time()
    
    abaqusfile = open(filename, 'r')
    
    nodeproxy = ElementProxy('node', 3)
    tetproxy = ElementProxy('tet', 4)
    triproxy = ElementProxy('tri', 3)

    line = abaqusfile.readline()
    # check we have the right kind of CUBIT output file here
    assert(line == '*HEADING\n')
    
    if ebs == None:
        include = True
    else:
        include = False
    currmap = None
    # read line one by one until the end
    while(line):
        # status check
        if (line.find('**', 0, 2) == 0):
            pass
        elif (line.find('*NODE', 0, 5) == 0):
            if (currmap != None):
                currmap.blockEnd()
            #print('Found *NODE section, start reading nodes.')
            lineset = line.split('=') 
            if (len(lineset) == 1):
                nodeproxy.blockBegin('AllNodes')
            else:
                nodeproxy.blockBegin(lineset[1].strip('\n'))
            currmap = nodeproxy
        elif (line.find('*ELEMENT', 0, 8) == 0):
            if (currmap != None):
                currmap.blockEnd()
            lineset = line.split(',') 
            elementtype = lineset[1].split('=')[1]
            blockname = 'AllElements'
            if (len(lineset) == 3):
                blockname = lineset[2].split('=')[1].strip('\n')
            if (ebs == None) or (blockname in ebs):
                include = True
                #print('Importing *Element section: %s.' % (blockname))
            else:
                include  = False
            if (elementtype == 'C3D4') and include:
                currmap = tetproxy
                tetproxy.blockBegin(blockname)
            elif (elementtype == 'STRI3') and include:
                currmap = triproxy
                triproxy.blockBegin(blockname)
        elif (line.find('**', 0, 2) == -1 and line[0] == '*'):
            if (currmap != None):
                currmap.blockEnd()
                currmap = None
        else:
            if (currmap == nodeproxy):
                linesec = line.replace(',', ' ')
                linesec = linesec.split() 
                nodeid = int(linesec[0])
                node = [0.0, 0.0, 0.0]
                # set to the right scale
                node[0] = float(linesec[1])*scale
                node[1] = float(linesec[2])*scale
                node[2] = float(linesec[3])*scale 
                currmap.insert(nodeid, node)
            elif (currmap == tetproxy) and include: 
                linesec = line.replace(',', ' ')
                linesec = linesec.split() 
                nodeid = int(linesec[0])
                node = [0,0,0,0]
                # set to the right scale
                node[0] = nodeproxy.getSTEPSID(int(linesec[1]))
                node[1] = nodeproxy.getSTEPSID(int(linesec[2]))
                node[2] = nodeproxy.getSTEPSID(int(linesec[3])) 
                node[3] = nodeproxy.getSTEPSID(int(linesec[4])) 
                currmap.insert(nodeid, node)
            elif (currmap == triproxy) and include:
                linesec = line.replace(',', ' ')
                linesec = linesec.split() 
                nodeid = int(linesec[0])
                node = [0,0,0]
                # set to the right scale
                node[0] = nodeproxy.getSTEPSID(int(linesec[1]))
                node[1] = nodeproxy.getSTEPSID(int(linesec[2]))
                node[2] = nodeproxy.getSTEPSID(int(linesec[3])) 
                currmap.insert(nodeid, node)           

        line = abaqusfile.readline()

    if (currmap != None):
        currmap.blockEnd()

    abaqusfile.close()
    print("Read Abaqus file succesfully")    
    
    nodedata = nodeproxy.getAllData()
    tetdata = tetproxy.getAllData() 
    tridata = triproxy.getAllData() 

    print("creating Tetmesh object in STEPS...")
    mesh = stetmesh.Tetmesh(nodedata, tetdata, tridata)
    print("Tetmesh object created.")
    
    return mesh, nodeproxy, tetproxy,triproxy

#############################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#############################################################################################

def importGmsh(filename, scale):
    """ 
    Read a Gmsh-formated mesh file, return the created steps.geom.Tetmesh object, 
    the element mapping for nodes, tetraedrons and triangles.
    
    PARAMETERS:
    
    * filename: the Abaqus filename (or path) including any suffix.
    * scale: LENGTH scale from the importing mesh to real geometry. e.g. a radius 
      of 10 in the importing file to a radius of 1 micron in STEPS, scale is 1e-7.
    
    RETURNS:
    
    mesh, nodeproxy, tetproxy, triproxy 

    * mesh: The STEPS TetMesh object
    * nodeproxy: Element Map for nodes
    * tetproxy: Element Map for tetrahedrons
    * triproxy: Element Map for triangles

   IMPORTANT NOTICE:
   
   User is recommanded to save the tetmesh objects using the saveTetmesh() method
   and recreate the objects from the saved files, instead of creating the objects
   via the importing functions each time, if the tetmesh objects are intented to
   be used in multiple simulations. Since the importing functions require a massive
   amount of time to create the Tetmesh object, comparing to the loadTetmesh() method.

    """
    print("Reading Gmsh file...")
    btime = time.time()
    
    meshfile = open(filename, 'r')
    
    nodeproxy = ElementProxy('node', 3)
    tetproxy = ElementProxy('tet', 4)
    triproxy = ElementProxy('tri', 3)

    line = meshfile.readline()
    assert(line == '$MeshFormat\n')
    line = meshfile.readline()
    line = meshfile.readline()
    assert(line == '$EndMeshFormat\n')
    line = meshfile.readline()
    assert(line == '$Nodes\n')
    line = meshfile.readline()
    n_nodes = int(line)
    # read nodes
    for i in range(n_nodes):
        line = meshfile.readline()
        linesec = line.split()
        id = int(linesec[0])
        node = [0.0,0.0,0.0]
        node[0] = float(linesec[1])*scale
        node[1] = float(linesec[2])*scale
        node[2] = float(linesec[3])*scale
        nodeproxy.insert(id, node)
    line = meshfile.readline()
    assert(line == '$EndNodes\n')   
    
    line = meshfile.readline()
    assert(line == '$Elements\n')
    
    line = meshfile.readline()
    n_elems = int(line)
    # read elem
    for i in range(n_elems):
        line = meshfile.readline()
        linesec = line.split()
        id = int(linesec[0])
        type = int(linesec[1])
        # triangle
        if type == 2:
            node = [0,0,0]
            node[0] = nodeproxy.getSTEPSID(int(linesec[6]))
            node[1] = nodeproxy.getSTEPSID(int(linesec[7]))
            node[2] = nodeproxy.getSTEPSID(int(linesec[8])) 
            triproxy.insert(id, node)
        # tet
        if type == 4:
            group_id = linesec[4]
            node = [0,0,0,0]
            node[0] = nodeproxy.getSTEPSID(int(linesec[6]))
            node[1] = nodeproxy.getSTEPSID(int(linesec[7]))
            node[2] = nodeproxy.getSTEPSID(int(linesec[8])) 
            node[3] = nodeproxy.getSTEPSID(int(linesec[9])) 
            tetproxy.insert(id, node)
            tetproxy.addToGroup(group_id, tetproxy.getSTEPSID(id))
            
    line = meshfile.readline()
    assert(line == '$EndElements\n')
    meshfile.close()
    print("Read Msh file succesfully")    
    
    nodedata = nodeproxy.getAllData()
    tetdata = tetproxy.getAllData() 
    tridata = triproxy.getAllData() 

    print("creating Tetmesh object in STEPS...")
    mesh = stetmesh.Tetmesh(nodedata, tetdata, tridata)
    print("Tetmesh object created.")
    
    return mesh, nodeproxy, tetproxy,triproxy



#############################################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#############################################################################################

def saveMesh(pathname, tetmesh):
    
    """ 
    Save a STEPS Tetmesh in two separate files 
    
    #. An XML file.
       This file stores the basic information about the mesh which tends to be
       common information for any software that supports tetrahedral meshes.
    
       * NODES are stored by cartesian coordinates.
       * TRIANGLES are stored by the indices of their 3 nodes.
       * TETRAHEDRONS are stored by the indices of their 4 nodes.
    
       The XML file also stores infomation about any compartments or 
       patches created in STEPS (class steps.geom.TmComp steps.geom.TmPatch
       respectively). 
      
       * COMPARTMENT(S) are stored by: 
       
         * their string identification.
         * a list of any volume systems added to the compartment at time of saving.
         * a list of tetrahedrons belonging to the compartment
        
       * PATCH(ES) are stored by:
       
         * their string identification.
         * a list of any surface systems added to the patch at time of saving.
         * the inner compartment id.
         * the outer compartment id (if it exists).
         * a list of trianlges belonging to this patch.
        
    #. An ASCII file storing information important to STEPS internally. This
       information must be found by STEPS once from the basic mesh infomation and
       is vital for simulations in STEPS. This can take a significant amount of 
       time for larger meshes, so storing this information in this way can drastically
       reduce future mesh loading times. The information stored is:
      
       * each triangle's area.
       * each triangle's normal.
       * each triangle's two (or one for surface tris) tetrahedron neighbours.
       * each tetrahedron's volume.
       * each tetrahedron's barycenter.
       * each tetrahedron's four neighbouring triangles.
       * each tetrahedron's four (or fewer for surface tets) tetrahedron neighbours.
    
    PARAMETERS:
    
    * pathname: 
      
      the root of the path to store the files. 
      
      e.g. 'meshes/spine1' will save data in /meshes/spine1.xml and /meshes/spine1.txt
      
    * tetmesh:
    
      A valid STEPS Tetmesh object (of class steps.geom.Tetmesh). 
      This mesh can be made in a variety of ways, e.g. to save a mesh loaded from tetgen::
      
          >>> import meshio
          >>> ### Use cubit script to create steps.geom.Tetmesh object from tetgen output file ###
          >>> mymesh = meshio.readTetgen(tetgenfilename)
          >>> ### Save this mesh in XML (and ASCII) format for quick-loading in future ###
          >>> meshio.saveMesh('/meshes/spine1', mymesh[0])
        
    """
    
    # Performa a basic test on the object itself
    if (tetmesh.__str__()[1:19] != 'steps.geom.Tetmesh'):
        print("2nd parameter not a valid steps.geom.Tetmesh object.")
        return 0
    
    # Following will throw IOError if pathname not a valid directory
    xmlfile = open(pathname+'.xml', 'w')
    textfile = open(pathname+'.txt', 'w')
    
    xmlfile.write('<?xml version="1.0" encoding="ISO-8859-1"?>\n')
    xmlfile.write('<tetmesh>\n')
    
    
    ### Write the node information  
    
    nverts = tetmesh.countVertices()
    
    xmlfile.write('\t<nodes size = "'+str(nverts)+'">\n')
    textfile.write(str(nverts)+'\n\n')
    
    for node in range(nverts):
        # Write indices and coordinates to xml file
        xmlfile.write('\t\t<node idx = "' + str(node) + '">\n')
        coords = str(tetmesh.getVertex(node)).strip(')').strip('(')
        xmlfile.write('\t\t\t<coords>'+ coords + '</coords>\n')
        xmlfile.write('\t\t</node>\n')
    xmlfile.write('\t</nodes>\n')
    
    
    ### Write the triangle information
    
    ntris = tetmesh.countTris()
    xmlfile.write('\t<triangles size = "' +str(ntris)+'">\n')
    textfile.write(str(ntris)+'\n')
    
    for tri in range(ntris):
        # Write indices and nodes to xml file
        xmlfile.write('\t\t<tri idx = "' + str(tri) + '">\n')
        nodes = str(tetmesh.getTri(tri)).strip(')').strip('(')
        xmlfile.write('\t\t\t<nodes>' + nodes + '</nodes>\n')
        xmlfile.write('\t\t</tri>\n')
        # Write the area, normal, tetneighbours to text file
        textfile.write(str(tetmesh.getTriArea(tri)) + " ")
        norm = tetmesh.getTriNorm(tri)
        textfile.write(str(norm[0]) + " " + str(norm[1]) + " " + str(norm[2]) + " ")
        tetneighb = tetmesh.getTriTetNeighb(tri)
        textfile.write(str(tetneighb[0]) + " " + str(tetneighb[1]) + " ")
        textfile.write('\n')
    xmlfile.write('\t</triangles>\n')
    
    
    ### Write the tetrahedron information
    
    ntets = tetmesh.countTets()
    xmlfile.write('\t<tetrahedrons size = "' +str(ntets)+'">\n')
    textfile.write('\n' +str(ntets)+'\n')
    
    for tet in range(ntets):
        # Write indices and nodes to xml file
        xmlfile.write ('\t\t<tet idx = "' + str(tet) + '">\n')
        nodes = str(tetmesh.getTet(tet)).strip(')').strip('(')
        xmlfile.write('\t\t\t<nodes>' + nodes + '</nodes>\n')
        xmlfile.write('\t\t</tet>\n')
        # Write the volume, barycenter, trineighbours, tetneighbours to xml file
        textfile.write(str(tetmesh.getTetVol(tet)) + " ")
        baryc = tetmesh.getTetBarycenter(tet)
        textfile.write(str(baryc[0]) + " " + str(baryc[1]) + " " + str(baryc[2]) + " ")
        trineighb = tetmesh.getTetTriNeighb(tet)
        textfile.write(str(trineighb[0])+" "+str(trineighb[1])+" "+str(trineighb[2])+" "+str(trineighb[3])+" ")
        tetneighb = tetmesh.getTetTetNeighb(tet)
        textfile.write(str(tetneighb[0])+" "+str(tetneighb[1])+" "+str(tetneighb[2])+" "+str(tetneighb[3])+" ")
        textfile.write('\n')
    xmlfile.write('\t</tetrahedrons>\n')
    
    
    ### Write the comp and patch information. 
    # TODO: Changes need to be made to steps code to make it easier to get the tet
    # and tri members. Currently the mesh returns base pointer (Comp or Patch) 
    # which cannot be used to find the indices. 
    
    comps = tetmesh.getAllComps()
    ncomps = comps.__len__()
    xmlfile.write('\t<compartments size = "' + str(ncomps) + '">\n')
    
    if (ncomps > 0) :
        ids = []
        vsys=[]
        tets = []
        for c in range(ncomps):
            ids.append(comps[c].getID())
            vsys.append(comps[c].getVolsys())
            tets.append([])
        assert(tets.__len__() == ncomps)
        # Only choice right now is to loop over all tets and compare comp to tet id
        for tet in range(ntets):
            comptemp = tetmesh.getTetComp(tet)
            if not comptemp: continue
            idtemp = comptemp.getID()
            for c in range(ncomps):
                if idtemp == ids[c]:
                    tets[c].append(tet)
                    break
        # Now we have the tet members of each comp, we can write this to xml
        for c in range(ncomps):
            xmlfile.write('\t\t<comp idx = "' +str(c) + '">\n')
            xmlfile.write('\t\t\t<id>' + ids[c] + '</id>\n')
            xmlfile.write('\t\t\t<volsys>')
            for v in vsys[c]:
                if (v != vsys[c][0]) : xmlfile.write(',')
                xmlfile.write(v)
            xmlfile.write('</volsys>\n')
            xmlfile.write('\t\t\t<tets>')
            for t in tets[c]:
                if (t != tets[c][0]) : xmlfile.write(',')
                xmlfile.write(str(t))
            xmlfile.write('</tets>\n')
            xmlfile.write('\t\t</comp>\n')
    xmlfile.write('\t</compartments>\n')
    
    patches = tetmesh.getAllPatches()
    npatches = patches.__len__()
    xmlfile.write('\t<patches size = "' + str(npatches) + '">\n')
    
    if (npatches > 0) :
        ids = []
        ssys = []
        icomp=[]
        ocomp=[]
        tris = []
        for p in range(npatches):
            ids.append(patches[p].getID())
            ssys.append(patches[p].getSurfsys())
            icomp.append(patches[p].getIComp().getID())
            if (not patches[p].getOComp()): ocomp.append('null')
            else : ocomp.append(patches[p].getOComp().getID())
            tris.append([])
        assert(ids.__len__() == ssys.__len__() == icomp.__len__() == ocomp.__len__() == tris.__len__() == npatches)
        
        for tri in range(ntris):
            patchtemp = tetmesh.getTriPatch(tri)
            if not patchtemp: continue
            idtemp = patchtemp.getID()
            for p in range(npatches):
                if idtemp == ids[p]:
                    tris[p].append(tri)
                    break
        # Write all to xml
        for p in range(npatches):
            xmlfile.write('\t\t<patch idx = "'+str(p) + '">\n')
            xmlfile.write('\t\t\t<id>' + ids[p] + '</id>\n')
            xmlfile.write('\t\t\t<surfsys>')
            for s in ssys[p]:
                if (s != ssys[p][0]) : xmlfile.write(',')
                xmlfile.write(s)
            xmlfile.write('</surfsys>\n')
            xmlfile.write('\t\t\t<icomp>' + icomp[p] + '</icomp>\n')
            xmlfile.write('\t\t\t<ocomp>' + ocomp[p] + '</ocomp>\n')
            xmlfile.write('\t\t\t<tris>')
            for t in tris[p]:
                if (t != tris[p][0] ) : xmlfile.write(',')
                xmlfile.write(str(t))
            xmlfile.write('</tris>\n')
            xmlfile.write('\t\t</patch>\n')
    xmlfile.write('\t</patches>\n')
    
    
    xmlfile.write('</tetmesh>')
    
    xmlfile.close()
    textfile.close()

#############################################################################################

def loadMesh(pathname):
    """ 
    Load a mesh in STEPS from the XML (and ASCII) file. This will
    work with just the XML file, but this is akin to creating the mesh in STEPS
    from scratch and really negates the use of storing the mesh infomation at all.
    For maximum benefit the XML file should be accompanied by the ASCII file, which
    contains all the internal information.
     
    PARAMETERS:
    
    * pathname: the root of the path where the file(s) are stored. 
      e.g. with 'meshes/spine1' this function will look for files /meshes/spine1.xml 
      and /meshes/spine1.txt
    
    RETURNS: A tuple (mesh, comps, patches)
    
    * mesh 
      The STEPS Tetmesh object (steps.geom.Tetmesh)
    * comps
      A list of any compartment objects (steps.geom.TmComp) from XML file 
    * patches
      A list of any patches objects (steps.geom.TmPatch) from XML file 
    """


    # Try to open the XML file. An error will be thrown if it doesn't exist
    xmlfile = open(pathname+'.xml', 'r')

    # Try to open the text file. A warning will be shown and a flag set if it doesn't exist 
    havetxt = True
    try :
        textfile = open(pathname+'.txt', 'r')
    except:
        havetxt = False
    if (havetxt == False) : print("WARNING: text file not found. Will construct mesh from information in XML file only.")
    
    # Perform a basic check to see if we have the expected kind of file which has not been altered.
    info = xmlfile.readline()
    if(xmlfile.readline().rstrip() != '<tetmesh>'):
        print('XML file is not a recognised STEPS mesh file')
        return
    
    # Collect basic node information and perform some checks on the data read from XML file
    nodeinfo = xmlfile.readline().strip()
    assert(nodeinfo.__len__() > 17)
    assert(nodeinfo[-2:] == '">')
    nnodes = int(nodeinfo[15:-2])
    if (havetxt): assert (nnodes == int(textfile.readline()))
    
    nodes_out = [0.0]*(nnodes*3)    
    for i in range(nnodes):
        idxtemp = xmlfile.readline().strip()
        assert(int(idxtemp[13:-2]) == i)
        coordtemp = xmlfile.readline().strip()
        assert(coordtemp[:8] == '<coords>' and coordtemp[-9:] == '</coords>')
        coordtemp = coordtemp[8:-9].split(', ')
        nodes_out[i*3], nodes_out[(i*3)+1], nodes_out[(i*3)+2] = float(coordtemp[0]), float(coordtemp[1]), float(coordtemp[2])
        assert(xmlfile.readline().strip() == '</node>')
    
    assert(xmlfile.readline().strip() == '</nodes>')
    
    # Now read triangle information from xml file and text file if we have it
    triinfo = xmlfile.readline().strip()
    assert(triinfo.__len__() > 21)
    assert(triinfo[-2:] == '">')
    ntris = int(triinfo[19:-2])
    if (havetxt) : 
        textfile.readline()
        assert (ntris == int(textfile.readline()))

    tris_out = [0]*(ntris*3) # numpy.zeros(ntris*3, dtype = 'int')
    if (havetxt) :
        triareas_out = [0.0]*ntris 		# numpy.zeros(ntris)
        trinorms_out = [0.0]*(ntris*3)	# numpy.zeros(ntris*3)
        tritetns_out = [0]*(ntris*2)	# numpy.zeros(ntris*2, dtype = 'int')
    
    for i in range(ntris): 
        idxtemp = xmlfile.readline().strip()
        assert(int(idxtemp[12:-2]) == i)
        nodetemp = xmlfile.readline().strip()
        assert (nodetemp[:7] == '<nodes>' and nodetemp[-8:] == '</nodes>')
        nodetemp = nodetemp[7:-8].split(', ')
        tris_out[i*3], tris_out[(i*3)+1], tris_out[(i*3)+2] = int(nodetemp[0]), int(nodetemp[1]), int(nodetemp[2])
        assert(xmlfile.readline().strip() == '</tri>')
        # Now read the text file, if it exists, and get the extra information
        if (havetxt):
            line = textfile.readline().rstrip().split(" ")
            assert(line.__len__() == 6)
            triareas_out[i], trinorms_out[i*3], trinorms_out[(i*3)+1], trinorms_out[(i*3)+2] = float(line[0]), float(line[1]), float(line[2]), float(line[3])
            tritetns_out[i*2], tritetns_out[(i*2)+1] = int(line[4]), int(line[5])       
    assert(xmlfile.readline().strip() == '</triangles>')
    
    # Now read tet information from xml file and text file if we have it
    tetinfo = xmlfile.readline().strip()
    assert(tetinfo.__len__() > 24)
    assert(tetinfo[-2:] == '">')
    ntets = int(tetinfo[22:-2])
    if (havetxt) :
        textfile.readline()
        assert(ntets == int(textfile.readline()))   
    
    tets_out = [0]*(ntets*4)	# numpy.zeros(ntets*4, dtype = 'int')
    if (havetxt): 
        tetvols_out =  [0.0]*ntets		# numpy.zeros(ntets)
        tetbarycs_out = [0.0]*(ntets*3)	# numpy.zeros(ntets*3)
        tettrins_out = 	[0]*(ntets*4)	# numpy.zeros(ntets*4, dtype = 'int')
        tettetns_out = 	[0]*(ntets*4)	# numpy.zeros(ntets*4, dtype = 'int')
    for i in range(ntets): 
        idxtemp = xmlfile.readline().strip()
        assert(int(idxtemp[12:-2]) == i)
        nodetemp = xmlfile.readline().strip()
        assert (nodetemp[:7] == '<nodes>' and nodetemp[-8:] == '</nodes>')      
        nodetemp = nodetemp[7:-8].split(', ')
        tets_out[i*4], tets_out[(i*4)+1], tets_out[(i*4)+2], tets_out[(i*4)+3] = int(nodetemp[0]), int(nodetemp[1]), int(nodetemp[2]), int(nodetemp[3])
        assert(xmlfile.readline().strip() == '</tet>')
        # Read the text file if we have it and get further information
        if (havetxt):
            line = textfile.readline().rstrip().split(" ")
            assert (line.__len__() == 12)
            tetvols_out[i], tetbarycs_out[i*3], tetbarycs_out[(i*3)+1], tetbarycs_out[(i*3)+2] = float(line[0]), float(line[1]), float(line[2]), float(line[3])
            tettrins_out[i*4], tettrins_out[(i*4)+1], tettrins_out[(i*4)+2], tettrins_out[(i*4)+3] = int(line[4]), int(line[5]), int(line[6]), int(line[7])
            tettetns_out[i*4], tettetns_out[(i*4)+1], tettetns_out[(i*4)+2], tettetns_out[(i*4)+3] = int(line[8]), int(line[9]), int(line[10]), int(line[11])
    assert(xmlfile.readline().strip() == '</tetrahedrons>')

    # We have all the information now. Time to make the Tetmesh object. New constructor keeps the order, which is:
    # nodes, tris, tri areas, tri normals, tri tet neighbours, tets, tet volumes, tet barycenters, tet tri neighbs, tet tet neighbs.
    mesh = stetmesh.Tetmesh(nodes_out, tris_out, triareas_out, trinorms_out, tritetns_out, tets_out, tetvols_out, tetbarycs_out, tettrins_out, tettetns_out)
    
    # Now fetch any comp and patch information from XML file
    compinfo = xmlfile.readline().strip()
    assert(compinfo.__len__() > 24)
    assert(compinfo[-2:] == '">')
    ncomps = int(compinfo[22:-2])
    comps_out = []
    for i in range(ncomps):
        idxtemp = xmlfile.readline().strip()
        assert(int(idxtemp[13:-2]) == i)
        idtemp = xmlfile.readline().strip()
        assert(idtemp[:4] == '<id>' and idtemp[-5:] == '</id>')
        idtemp = idtemp[4:-5]
        volsystemp = xmlfile.readline().strip()
        assert(volsystemp[:8] == '<volsys>' and volsystemp[-9:] == '</volsys>')
        volsystemp = volsystemp[8:-9].split(',')
        if (volsystemp[0] == '') : volsystemp = []
        tettemp = xmlfile.readline().strip()
        assert(tettemp[:6] == '<tets>' and tettemp[-7:] == '</tets>')
        tettemp = tettemp[6:-7].split(',')
        nctets = tettemp.__len__()
        ctets = [0]*nctets		# numpy.zeros(nctets, dtype = 'int')
        for ct in range(nctets): ctets[ct] = int(tettemp[ct])
        c_out = stetmesh.TmComp(idtemp, mesh, ctets)
        for v in volsystemp: c_out.addVolsys(v)
        comps_out.append(c_out) 
        assert(xmlfile.readline().strip() == '</comp>')
    assert(xmlfile.readline().strip() == '</compartments>')
    
    # Retrieve patch info
    patchinfo = xmlfile.readline().strip()
    assert(patchinfo.__len__() > 19)
    assert(patchinfo[-2:] == '">')
    npatches = int(patchinfo[17:-2])
    patches_out = []
    for i in range(npatches):
        idxtemp = xmlfile.readline().strip()
        assert(int(idxtemp[14:-2]) == i)
        idtemp = xmlfile.readline().strip()
        assert(idtemp[:4] == '<id>' and idtemp[-5:] == '</id>')
        idtemp = idtemp[4:-5]
        surfsystemp = xmlfile.readline().strip()
        assert(surfsystemp[:9] == '<surfsys>' and surfsystemp[-10:] == '</surfsys>')
        surfsystemp = surfsystemp[9:-10].split(',')
        if (surfsystemp[0] == '') : surfsystemp = []
        icomptemp = xmlfile.readline().strip()
        assert(icomptemp[:7] == '<icomp>' and icomptemp[-8:] == '</icomp>')
        icomptemp = icomptemp[7:-8]
        ocomptemp = xmlfile.readline().strip()
        assert(ocomptemp[:7] == '<ocomp>' and ocomptemp[-8:] == '</ocomp>')
        ocomptemp = ocomptemp[7:-8]     
        tritemp = xmlfile.readline().strip()
        assert(tritemp[:6] == '<tris>' and tritemp[-7:] == '</tris>')       
        tritemp = tritemp[6:-7].split(',')
        nptris = tritemp.__len__()
        ptris = [0]*nptris		# numpy.zeros(nptris, dtype='int')
        for pt in range(nptris): ptris[pt] = int(tritemp[pt])
        if (ocomptemp != 'null'): p_out = stetmesh.TmPatch(idtemp, mesh, ptris, mesh.getComp(icomptemp), mesh.getComp(ocomptemp))
        else :  p_out = stetmesh.TmPatch(idtemp, mesh, ptris, mesh.getComp(icomptemp))
        for s in surfsystemp: p_out.addSurfsys(s)
        patches_out.append(p_out)
        assert(xmlfile.readline().strip() == '</patch>')
    assert(xmlfile.readline().strip() == '</patches>')

    
    return (mesh,comps_out,patches_out)




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

failed_tests=[]

# # # # # # # # # # # # # # # REACTIONS # # # # # # # # # # # # # # # # #

print "\n1: Reaction - First order, irreversible:"
import first_order_irev
if first_order_irev.passed: print "PASSED\n"
else: 
    print "FAILED\n"
    failed_tests.append("1: Reaction - First order, irreversible")

print "2: Reaction - First order, reversible:"
import first_order_rev
if first_order_rev.passed: print "PASSED\n"
else: 
    print "FAILED\n"
    failed_tests.append("2: Reaction - First order, reversible")

print "3: Reaction - Second order, irreversible, A0=B0:"
import second_order_irev_AA
if second_order_irev_AA.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("3: Reaction - Second order, irreversible, A0=B0")

print "4: Reaction - Second order, irreversible, A0!=B0:"
import second_order_irev_AB
if second_order_irev_AB.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("4: Reaction - Second order, irreversible, A0!=B0")

print "5: Reaction - Second-order, irreversible, 2D:"
import second_order_irev_2D
if second_order_irev_2D.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("5: Reaction - Second order, irreversible, 2D")

print "6: Reaction - Production and degradation:"
import masteq
if masteq.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("6: Reaction - Production and degradation")
    num_failed = 1

# # # # # # # # # # # # # # # DIFFUSION # # # # # # # # # # # # # # # # #

print "7: Diffusion - Unbounded:"
import unbdiff
if unbdiff.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("7: Diffusion - Unbounded")
    num_failed = 1

print "8: Diffusion - Bounded:"
import bounddiff
if bounddiff.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("8: Diffusion - Bounded")
    num_failed = 1

print "9: Diffusion - Clamped:"
import csd_clamp
if csd_clamp.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("9: Diffusion - Clamped")
    num_failed = 1

# # # # # # # # # # # # # #  REACTION-DIFFUSION # # # # # # # # # # # # # 

print "10: Reaction-diffusion - Degradation-diffusion:"
import kisilevich
if kisilevich.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("10: Reaction-diffusion - Degradation-diffusion")
    num_failed = 1

print "11: Reaction-diffusion - Production and second order degradation:"
import masteq_diff
if masteq_diff.passed: print "PASSED\n"
else:
    print "FAILED\n"
    failed_tests.append("11: Reaction-diffusion - Production and second order degradation")
    num_failed = 1

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


print ""
print len(failed_tests), "tests failed:"
for t in failed_tests:
    print t, 



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from . import steps_swig
import _steps_swig

### Now defunct mesh saving/loading tool ###
# from steps_swig import loadASCII, saveASCII


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

class RNG(steps_swig.RNG):
    """
        Base class for all random number generators in STEPS.
    """
    def __init__(self, *args): 
        """
        """
        this = _steps_swig.new_RNG(*args)
        try: self.this.append(this)
        except: self.this = this
        self.thisown = True

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

def create(*args):
  """
    Creates and returns a reference to a steps.rng.RNG random number generator object, 
    which is specified by type and pre-allocates a buffer list with size of buffer_size.

    Syntax::
        
        create(type, buffer_size)

    Arguments:
        * string type
        * uint buffer_size

    Return:
        steps.rng.RNG

    """
  return steps_swig.create(*args)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

def create_mt19937(*args):
  """
    Creates and returns a reference to a steps.rng.RNG random number generator object, 
    which is specified by type and pre-allocates a buffer list with size of buffer_size.

    Syntax::
        
        create_mt19937(buffer_size)

    Arguments:
        * uint buffer_size

    Return:
        steps.rng.RNG

    """
  return steps_swig.create_mt19937(*args)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#  Last Changed Rev:  $Rev: 423 $
#  Last Changed Date: $Date: 2011-06-02 10:10:54 +0900 (Thu, 02 Jun 2011) $
#  Last Changed By:   $Author: iain $

try:
    from setuptools import setup, Extension
    
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension


def name():
    return 'STEPS'
    
def version():
    return '1.2.1'
    
def author():
    return 'STEPS Development Team'
    
def email():
    return 'steps.dev@gmail.com'
    
def url():
    return 'http://steps.sourceforge.net'
    
def desc():
    return 'STochastic Engine for Pathway Simulation'

def download():
    return 'http://sourceforge.net/projects/steps/files/src'

def platforms():
    return ['Mac OS X', 'Windows XP', 'Windows Vista', 'Linux', 'Unix']
    
def license():
    return 'GNU General Public License Version 3.0'
    
def packages():
    return ['steps', 'steps/utilities']
  
def steps_ext():
    return dict(
        name='_steps_swig',
        
        sources=['cpp/error.cpp',
        'cpp/model/model.cpp', 'cpp/model/diff.cpp',
        'cpp/model/reac.cpp','cpp/model/spec.cpp','cpp/model/sreac.cpp',
        'cpp/model/surfsys.cpp','cpp/model/volsys.cpp',
        
        'cpp/math/tetrahedron.cpp','cpp/geom/tetmesh.cpp',
        'cpp/math/linsolve.cpp','cpp/math/triangle.cpp',
        'cpp/geom/comp.cpp','cpp/geom/geom.cpp','cpp/geom/patch.cpp',
        'cpp/geom/tet.cpp', 'cpp/geom/tetmesh_rw.cpp',
        'cpp/geom/tmcomp.cpp','cpp/geom/tmpatch.cpp', 'cpp/geom/diffboundary.cpp',
        'cpp/geom/tri.cpp',

        
        'cpp/math/tools.cpp',
        'cpp/rng/rng.cpp', 'cpp/rng/mt19937.cpp',

        'cpp/solver/api_comp.cpp','cpp/solver/api_main.cpp',
        'cpp/solver/api_patch.cpp','cpp/solver/api_tet.cpp',
        'cpp/solver/api_tri.cpp', 'cpp/solver/api_diffboundary.cpp',
        'cpp/solver/compdef.cpp',
        'cpp/solver/diffdef.cpp','cpp/solver/patchdef.cpp',
        'cpp/solver/reacdef.cpp','cpp/solver/specdef.cpp',
        'cpp/solver/sreacdef.cpp', 'cpp/solver/diffboundarydef.cpp',
        'cpp/solver/statedef.cpp',
        
        'cpp/tetexact/comp.cpp','cpp/tetexact/diff.cpp',
        'cpp/tetexact/kproc.cpp','cpp/tetexact/patch.cpp',
        'cpp/tetexact/reac.cpp','cpp/tetexact/sreac.cpp',
        'cpp/tetexact/tet.cpp','cpp/tetexact/tetexact.cpp',
        'cpp/tetexact/tri.cpp', 'cpp/tetexact/diffboundary.cpp',
        
        'cpp/wmdirect/comp.cpp','cpp/wmdirect/kproc.cpp',
        'cpp/wmdirect/patch.cpp','cpp/wmdirect/reac.cpp',
        'cpp/wmdirect/sreac.cpp','cpp/wmdirect/wmdirect.cpp',
        
        'cpp/wmrk4/wmrk4.cpp',
        
        'swig/steps_wrap.cpp'],
        
        undef_macros=['NDEBUG']
    )
        
        
def ext_modules():
    modules = [steps_ext()]
    return modules

ExtModule = lambda extension:  Extension(**extension)

setup(name = name(),
      version = version(),
      author = author(),
      author_email = email(),
      url = url(),
      description = desc(),
      download_url = download(),
      platforms = platforms(),
      license = license(),
      
      packages = packages(), 
      
      ext_package = 'steps',
      
      ext_modules  = [ExtModule(ext) for ext in ext_modules()]
      )





# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import math
import numpy
import time 
from pylab import *

import steps.model as smod
import steps.geom as sgeom
import steps.rng as srng
import steps.solver as ssolv
import steps.utilities.meshio as meshio

from tol_funcs import *

### NOW   A+B-> B,  0->A (see Erban and Chapman, 2009)

########################################################################
SCALE = 1.0

KCST_f = 100e6*SCALE			# The reaction constant, degradation
KCST_b = (20.0e-10*SCALE) 		# The reaction constant, production

DCST_A = 20e-12
DCST_B = 20e-12

B0 = 1   # The number of B moleucles

DT = 0.1			# Sampling time-step
INT = 50000.1 		# Sim endtime

filename = 'cube_1_1_1_73tets.inp'

# A tolerance of 7.5% will fail <1% of the time
tolerance = 7.5/100

########################################################################

mdl  = smod.Model()

A = smod.Spec('A', mdl)
B = smod.Spec('B', mdl)

volsys = smod.Volsys('vsys',mdl)

diffA = smod.Diff('diffA', volsys, A, dcst = DCST_A)
diffB = smod.Diff('diffB', volsys, B, dcst = DCST_B)

# Production
R1 = smod.Reac('R1', volsys, lhs = [A, B], rhs = [B], kcst = KCST_f)
R2 = smod.Reac('R2', volsys, lhs = [], rhs = [A], kcst = KCST_b)

geom = meshio.loadMesh('validation/meshes/'+filename)[0]

comp1 = sgeom.TmComp('comp1', geom, range(geom.ntets))
comp1.addVolsys('vsys')

rng = srng.create('mt19937', 512)
rng.initialize(int(time.time()%4294967295))

sim = ssolv.Tetexact(mdl, geom, rng)
sim.reset()

tpnts = numpy.arange(0.0, INT, DT)
ntpnts = tpnts.shape[0]

res = numpy.zeros([ntpnts])
res_std1 = numpy.zeros([ntpnts])
res_std2 = numpy.zeros([ntpnts])

sim.reset()
sim.setCompCount('comp1', 'A', 0)
sim.setCompCount('comp1', 'B', B0)

b_time = time.time()
for t in range(0, ntpnts):
    sim.run(tpnts[t])
    res[t] = sim.getCompCount('comp1', 'A')


def fact(x): return (1 if x==0 else x * fact(x-1))

# Do cumulative count, but not comparing them all. 
# Don't get over 50 (I hope)
steps_n_res = numpy.zeros(50)
for r in res: steps_n_res[r]+=1
for s in range(50): steps_n_res[s] = steps_n_res[s]/ntpnts

passed= True
max_err = 0.0

k1 = KCST_f/6.022e23
k2 = KCST_b*6.022e23
v = comp1.getVol()*1.0e3 # litres

for m in range(5, 11):
    analy = (1.0/fact(m))*math.pow((k2*v*v)/(B0*k1), m)*math.exp(-((k2*v*v)/(k1*B0)))
    if not tolerable(steps_n_res[m], analy, tolerance):
        passed = False
    #if (abs(2*(steps_n_res[m]-analy)/(steps_n_res[m]+analy)) > max_err): max_err = abs(2*(steps_n_res[m]-analy)/(steps_n_res[m]+analy))
#print "Max error: ", max_err*100, "%"





# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STEPS - STochastic Engine for Pathway Simulation
# Copyright (C) 2007-2011 Okinawa Institute of Science and Technology, Japan.
# Copyright (C) 2003-2006 University of Antwerp, Belgium.
#
# See the file AUTHORS for details.
#
# This file is part of STEPS.
#
# STEPS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# STEPS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

# Example: SBML import
# http://steps.sourceforge.net/manual/sbml_importer.html

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import steps.rng as srng
import steps.solver as ssolver
import steps.utilities.sbml as ssbml
import numpy
from pylab import *

def runSBMLmod(sbmlFile, time_end, time_dt, iter_n = 1, solver = 'Wmdirect', specs_plot = {}, vol_units = 1.0e-3, vol_def = False):
    iSbml = ssbml.Interface(sbmlFile, volunits_def = vol_units, volume_def = vol_def )
    mdl = iSbml.getModel()
    mesh = iSbml.getGeom()
    comp_specs = {}
    if not specs_plot: got_sp = False
    else: got_sp = True
    for comp in mesh.getAllComps():
        comp_specs[comp.getID()] = []
        volsys_strings = comp.getVolsys()
        for vsys_str in volsys_strings:
            vsys = mdl.getVolsys(vsys_str)
            specs = vsys.getAllSpecs()
            for spec in specs:
                comp_specs[comp.getID()].append(spec.getID())
                if (got_sp == False): specs_plot.update({spec.getID():''})
    comp_specs_n = 0
    for comp in comp_specs:
        comp_specs[comp].sort()
        comp_specs_n += len(comp_specs[comp])
    time_pnts = numpy.arange(0.0, time_end, time_dt)
    points_n = int(round(time_end/time_dt))
    if (len(time_pnts)  == (points_n + 1)): time_pnts = numpy.delete(time_pnts, len(time_pnts)-1)
    res = numpy.zeros([iter_n, points_n, comp_specs_n])
    r = srng.create('mt19937', 256)
    r.initialize(7233)
    if (solver == 'Wmdirect'):
        sim = ssolver.Wmdirect(mdl, mesh, r)
    elif (solver == 'Wmrk4'):
        sim = ssolver.Wmrk4(mdl, mesh, r)
        sim.setDT(0.0001)
    else:
        raise NameError("Unsupported solver. SBML importer accepts well-mixed solvers 'Wmrk4' and 'Wmdirect'")
    for it in range (0, iter_n):
        sim.reset()
        iSbml.reset()
        iSbml.setupSim(sim)
        for tp in range(0, points_n):
            sim.run(time_pnts[tp])
            i = 0
            for comp in comp_specs:
                for spec in comp_specs[comp]:
                    res[it, tp, i] = sim.getCompConc(comp, spec)
                    i+=1
            iSbml.updateSim(sim, time_dt)
    mean_res = numpy.mean(res, 0)
    i=0
    for comp in comp_specs:
        for spec in comp_specs[comp]:
            if (spec in specs_plot):
                if (specs_plot[spec]): plot(time_pnts, mean_res[:,i], label = spec+ ", "+ comp, color = specs_plot[spec])
                else: plot(time_pnts, mean_res[:,i], label = spec+ ", "+ comp)
            i+=1
    legend(loc = 'best', numpoints=1)
    xlabel('Time (s)')
    ylabel('Conc (M)')
    show()

print "Running Biomodel 98 in deterministic solver..."
runSBMLmod('biomodel/BIOMD0000000098.xml', 10, 0.001, solver='Wmrk4',specs_plot={'Y':'blue', 'Z':'red'})

print "\nRunning Biomodel 98 in well-mixed stochastic solver in reduced volume..."
runSBMLmod('biomodel/BIOMD0000000098.xml', 10, 0.001, solver='Wmdirect',specs_plot={'Y':'blue', 'Z':'red'}, vol_units = 1.0e-3, vol_def = 1.0e-18)

